import operator
from functools import reduce

from django.core.exceptions import FieldDoesNotExist
from django.db.models import QuerySet, Q
from django.db.models.manager import Manager

from .fields import PathField


def _get_path_fields(model, name=None):
    if name is None:
        return [f for f in model._meta.fields if isinstance(f, PathField)]
    return [model._meta.get_field(name)]


def _get_path_field(model, name):
    path_fields = _get_path_fields(model, name)
    n = len(path_fields)
    if n == 0:
        raise FieldDoesNotExist(
            'A `PathField` needs to be defined '
            'in order to use `TreeModelMixin`.')
    if n == 1:
        return path_fields[0]
    raise ValueError(
        'You need to specify which `PathField` to use for this query '
        'among these values: %s' % [f.name for f in path_fields])


# TODO: Implement a faster `QuerySet.delete` and add it to the benchmark.
class TreeQuerySetMixin:
    def _get_path_field_attname(self, name):
        return _get_path_field(self.model, name).attname

    def filter_roots(self, path_field=None):
        attname = self._get_path_field_attname(path_field)
        return self.filter(**{f'{attname}__level': 1})

    def get_descendants(self, include_self=False, path_field=None):
        attname = self._get_path_field_attname(path_field)
        # TODO: Avoid doing an extra query.
        ancestor_paths = list(self.values_list(attname, flat=True))
        queryset = self.model.objects.all()
        if not ancestor_paths:
            return queryset.none()
        if not include_self:
            queryset = queryset.exclude(**{attname + '__in': ancestor_paths})
        return queryset.filter(
            reduce(operator.or_, [
                Q(**{attname + '__descendant_of': ancestor_path})
                for ancestor_path in ancestor_paths
            ])
        )


class TreeQuerySet(TreeQuerySetMixin, QuerySet):
    pass


class TreeManager(Manager.from_queryset(TreeQuerySet)):
    pass



import re
from typing import List, Optional


# TODO: Integrate the doctests to the test runner.


UNNECESSARY_QUOTE_RE = re.compile(r'^[a-z_]+$')


def quote_ident(identifier: str):
    """
    >>> quote_ident('usual')
    'usual'
    >>> quote_ident('usual_with_underscores')
    'usual_with_underscores'
    >>> quote_ident('column with spaces')
    '"column with spaces"'
    >>> quote_ident('quote"inside')
    '"quote""inside"'
    >>> quote_ident('Capitalized')
    '"Capitalized"'
    >>> quote_ident('_private')
    '_private'
    >>> quote_ident('a')
    'a'
    >>> quote_ident('€')
    '"€"'
    >>> quote_ident('prénom')
    '"prénom"'
    >>> quote_ident('"already quoted"')
    '\"""already quoted""\"'
    """
    if UNNECESSARY_QUOTE_RE.match(identifier) is not None:
        return identifier
    identifier = identifier.replace('"', '""')
    return f'"{identifier}"'


def join_or(expressions: List[str]):
    if len(expressions) == 1:
        return expressions[0]
    return f'({" OR ".join(expressions)})'


def join_and(expressions: List[str]):
    return ' AND '.join(expressions)


def compare_columns(
    left: str, right: str, greater: Optional[bool] = None,
    strict: bool = False, nulls_last: bool = True,
):
    """
    >>> compare_columns('name', 'NEW.name')
    '(name IS NULL AND NEW.name IS NULL OR coalesce(name = NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', strict=True)
    '(name IS NULL AND NEW.name IS NULL OR coalesce(name = NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', nulls_last=False)
    '(name IS NULL AND NEW.name IS NULL OR coalesce(name = NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=True)
    '(name IS NULL OR coalesce(name >= NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=True, strict=True)
    '(name IS NULL AND NEW.name IS NOT NULL OR coalesce(name > NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=True, strict=True, nulls_last=False)
    '(NEW.name IS NULL AND name IS NOT NULL OR coalesce(name > NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=False)
    '(NEW.name IS NULL OR coalesce(name <= NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=False, strict=True)
    '(NEW.name IS NULL AND name IS NOT NULL OR coalesce(name < NEW.name, FALSE))'
    >>> compare_columns('name', 'NEW.name', greater=False, nulls_last=False)
    '(name IS NULL OR coalesce(name <= NEW.name, FALSE))'
    """
    operator = {
        None: '=',
        True: '>' if strict else '>=',
        False: '<' if strict else '<=',
    }[greater]
    operation = f'coalesce({left} {operator} {right}, FALSE)'
    if operator == '=':
        null_condition = join_and([f'{left} IS NULL', f'{right} IS NULL'])
    else:
        null_on_right = not nulls_last if greater else nulls_last
        null_condition = (
            f'{right} IS NULL' if null_on_right
            else f'{left} IS NULL'
        )
        if strict:
            null_condition = join_and([
                null_condition,
                f'{left} IS NOT NULL' if null_on_right
                else f'{right} IS NOT NULL'
            ])
    return join_or([null_condition, operation])


def get_nearby_sibling_where_clause(
    columns_in_order: List[str],
    record_name: str,
    greater: bool = True,
    nulls_last: bool = True,
):
    """
    >>> get_nearby_sibling_where_clause(["col1"], 'NEW')
    '(col1 IS NULL OR coalesce(col1 >= NEW.col1, FALSE))'
    >>> get_nearby_sibling_where_clause(["col1"], 'NEW', greater=False)
    '(NEW.col1 IS NULL OR coalesce(col1 <= NEW.col1, FALSE))'
    >>> get_nearby_sibling_where_clause(["col1", "col2", "col3"], 'NEW', greater=True)
    '((col1 IS NULL AND NEW.col1 IS NOT NULL OR coalesce(col1 > NEW.col1, FALSE)) OR (col1 IS NULL AND NEW.col1 IS NULL OR coalesce(col1 = NEW.col1, FALSE)) AND (col2 IS NULL AND NEW.col2 IS NOT NULL OR coalesce(col2 > NEW.col2, FALSE)) OR (col1 IS NULL AND NEW.col1 IS NULL OR coalesce(col1 = NEW.col1, FALSE)) AND (col2 IS NULL AND NEW.col2 IS NULL OR coalesce(col2 = NEW.col2, FALSE)) AND (col3 IS NULL OR coalesce(col3 >= NEW.col3, FALSE)))'
    """
    return join_or([
        join_and([
            compare_columns(
                column,
                f'{record_name}.{column}',
                greater=greater if i == column_index - 1 else None,
                strict=column_index < len(columns_in_order),
                nulls_last=nulls_last,
            )
            for i, column in enumerate(columns_in_order[:column_index])
        ])
        for column_index in range(1, len(columns_in_order) + 1)
    ])


def get_prev_sibling_where_clause(
    columns_in_order: List[str], record_name: str,
):
    return get_nearby_sibling_where_clause(
        columns_in_order=columns_in_order,
        record_name=record_name,
        greater=False,
    )


def get_next_sibling_where_clause(
    columns_in_order: List[str], record_name: str,
):
    return get_nearby_sibling_where_clause(
        columns_in_order=columns_in_order,
        record_name=record_name,
        greater=True,
    )



from random import choice
from string import ascii_letters

from django.db.models import CharField, ForeignKey, CASCADE
from treebeard.mp_tree import MP_Node
from treebeard.ns_tree import NS_Node
from treebeard.al_tree import AL_Node
from tree.fields import PathField
from tree.models import TreeModel
from mptt.models import MPTTModel, TreeForeignKey


def get_random_name():
    return ''.join([choice(ascii_letters) for _ in range(7)])


class MPTTPlace(MPTTModel):
    name = CharField(max_length=50, unique=True, default=get_random_name)
    parent = TreeForeignKey('self', null=True, blank=True,
                            related_name='children', on_delete=CASCADE)

    class MPTTMeta:
        order_insertion_by = ('name',)


class TreePlace(TreeModel):
    name = CharField(max_length=50, unique=True, default=get_random_name)
    parent = ForeignKey('self', null=True, blank=True, related_name='children',
                        on_delete=CASCADE)
    path = PathField(order_by=['name'], db_index=True)

    class Meta:
        indexes = [
            *PathField.get_indexes('treeplace', 'path'),
        ]


class TreebeardALPlace(AL_Node):
    name = CharField(max_length=50, unique=True, default=get_random_name)
    parent = ForeignKey('self', null=True, blank=True, related_name='children',
                        on_delete=CASCADE)
    node_order_by = ('name',)


class TreebeardMPPlace(MP_Node):
    name = CharField(max_length=50, unique=True, default=get_random_name)
    node_order_by = ('name',)


class TreebeardNSPlace(NS_Node):
    name = CharField(max_length=50, unique=True, default=get_random_name)
    node_order_by = ('name',)



from math import log10


class SkipTest(Exception):
    pass


SI_PREFIXES = (
    (1e9, 'G'),
    (1e6, 'M'),
    (1e3, 'k'),
    (1, ''),
    (1e-3, 'm'),
    (1e-6, 'µ'),
    (1e-9, 'n'),
)


def get_precision(n):
    return -int(log10(n))


def prefix_unit(v, unit, min_limit=None):
    if v is None:
        return

    prefixes = SI_PREFIXES
    if min_limit is not None:
        prefixes = prefixes[:min_limit]

    precision = get_precision(min([n for n, s in prefixes]))

    for exp, exp_str in prefixes:
        if v >= exp:
            break

    n = v / exp
    pat = ('%%.%df' % (precision - get_precision(exp)))
    res = '%s' % (pat % n)

    # We remove trailing zero and the dot if it's possible.
    if '.' in res:
        res = res.rstrip('0')
        if res[-1] == '.':
            res = res[:-1]

    if res == '0':
        return res

    return '%s %s%s' % (res, exp_str, unit)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('tree', '0001_initial')
    ]

    operations = [
        migrations.RunSQL("""
            DROP FUNCTION IF EXISTS rebuild_paths(
                table_name text, pk text, parent text, path text
            ) CASCADE;
            DROP FUNCTION IF EXISTS update_paths() CASCADE;
            DROP FUNCTION IF EXISTS from_alphanum(label text) CASCADE;
            DROP FUNCTION IF EXISTS to_alphanum(i bigint, size smallint) CASCADE;
        """),
    ]



DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'tree',
        'USER': 'tree',
    },
}

INSTALLED_APPS = (
    'tree',
    'mptt',
    'treebeard',

    'benchmark',
)

SECRET_KEY = 'not important here'

DATABASE_ROUTERS = ('benchmark.router.AutoRouter',)



from django.db.models import Lookup


class AncestorOf(Lookup):
    lookup_name = 'ancestor_of'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return (
            '%s = (%s)[:array_length(%s, 1)]' % (lhs, rhs, lhs),
            lhs_params + rhs_params + lhs_params,
        )


class SiblingOf(Lookup):
    lookup_name = 'sibling_of'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        # TODO: Simplify using `trim_array` once support for PostgreSQL < 14
        #       is dropped.
        return (
            '(%s)[:array_length(%s, 1) - 1] = (%s)[:array_length(%s, 1) - 1]'
            % (lhs, lhs, rhs, rhs),
            lhs_params + lhs_params + rhs_params + rhs_params,
        )


class ChildOf(Lookup):
    lookup_name = 'child_of'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        # TODO: Simplify using `trim_array` once support for PostgreSQL < 14
        #       is dropped.
        return (
            '(%s)[:array_length(%s, 1) - 1] = %s' % (lhs, lhs, rhs),
            lhs_params + lhs_params + rhs_params
        )


class DescendantOf(Lookup):
    lookup_name = 'descendant_of'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return (
            '(%s)[:array_length(%s, 1)] = %s' % (lhs, rhs, rhs),
            lhs_params + rhs_params + rhs_params,
        )



from django.db import migrations
from tree.fields import PathField
from tree.operations import (
    CreateTreeTrigger, RebuildPaths, DeleteTreeTrigger,
)


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0002_add_tmp_model'),
    ]

    operations = [
        migrations.AddField('Something', 'path', PathField(order_by=['name'])),
        CreateTreeTrigger('Something'),
        RebuildPaths('Something'),
        migrations.AlterField('Something', 'path', PathField(order_by=['name'])),
        DeleteTreeTrigger('Something'),
        migrations.DeleteModel('Something'),
    ]



from django.db import models, migrations
from django.db.models import CASCADE


def populate(apps, schema_editor):
    Something = apps.get_model('tests.Something')
    Something.objects.bulk_create([Something(name='%d' % i)
                                   for i in range(10)])
    bulk = []
    for i in range(10):
        parent = Something.objects.order_by('?')[0]
        bulk.append(Something(name='%s > %d' % (parent.name, i),
                              parent=parent))
    Something.objects.bulk_create(bulk)


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Something',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50)),
                ('parent', models.ForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
            ],
        ),
        migrations.RunPython(populate),
    ]



from typing import Type

from django.db.models import Model
from django.db.models.signals import post_save
from django.dispatch import receiver

from tree.fields import PathField


@receiver(post_save)
def defer_paths(sender: Type[Model], **kwargs):
    path_fields = [
        field for field in sender._meta.concrete_fields
        if isinstance(field, PathField)
    ]
    if path_fields:
        instance = kwargs['instance']
        for path_field in path_fields:
            if path_field.attname in instance.__dict__:
                # Removes the cached value for the field, making it deferred.
                # That way, Django will run a new query to know what is
                # the new path, only if it is used.
                # I wish we could make Django receive paths from SQL
                # through `RETURNING`, but unfortunately the ORM
                # only uses `RETURNING pk`.
                del instance.__dict__[path_field.attname]



from django.db import models, migrations
from django.db.models import CASCADE, Index, F
from django.db.models.expressions import RawSQL

from tree.fields import PathField
from tree.models import TreeModelMixin
from tree.operations import CreateTreeTrigger


class Migration(migrations.Migration):
    dependencies = [
        ('tree', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Place',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50)),
                ('parent', models.ForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
                ('path', PathField(order_by=['name'])),
            ],
            options={
                'ordering': ['path', 'name'],
                'indexes': [
                    Index(RawSQL('path[:array_length(path, 1) - 1]', ()), name='place_path_parent_index'),
                    Index(F('path__level'), name='place_path_level_index'),
                    Index(F('path__0_1'), name='place_path_slice_1_index'),
                    Index(F('path__0_2'), name='place_path_slice_2_index'),
                    Index(F('path__0_3'), name='place_path_slice_3_index'),
                    Index(F('path__0_4'), name='place_path_slice_4_index'),
                    Index(F('path__0_5'), name='place_path_slice_5_index'),
                ],
            },
            bases=(TreeModelMixin, models.Model),
        ),
        CreateTreeTrigger('tests.Place'),
        migrations.CreateModel(
            name='Person',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('century', models.SmallIntegerField(null=True, blank=True)),
                ('first_name', models.CharField(blank=True, max_length=20)),
                ('last_name', models.CharField(max_length=50)),
                ('parent', models.ForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
                ('path', PathField(order_by=['century', 'last_name', 'first_name'])),
            ],
            options={
                'ordering': ['path'],
                'indexes': [
                    Index(RawSQL('path[:array_length(path, 1) - 1]', ()), name='person_path_parent_index'),
                    Index(F('path__len'), name='person_path_length_index'),
                    Index(F('path__0_1'), name='person_path_slice_1_index'),
                    Index(F('path__0_2'), name='person_path_slice_2_index'),
                    Index(F('path__0_3'), name='person_path_slice_3_index'),
                    Index(F('path__0_4'), name='person_path_slice_4_index'),
                    Index(F('path__0_5'), name='person_path_slice_5_index'),
                ],
            },
            bases=(TreeModelMixin, models.Model),
        ),
        CreateTreeTrigger('tests.Person'),
    ]



from typing import List, Optional, Type

from django.db import DEFAULT_DB_ALIAS, connections
from django.db.models import Model

from .base import (
    quote_ident, get_prev_sibling_where_clause, get_next_sibling_where_clause,
    compare_columns, join_and,
)


def execute_format(
    sql: str,
    *args: str,
    using: Optional[List[str]] = None,
    into: Optional[List[str]] = None,
):
    if using is None:
        using = []
    if into is None:
        into = []

    sql = sql.replace("'", "''")

    args = ', ' + (', '.join(args)) if args else ''
    using = ' USING ' + ', '.join(using) if using else ''
    into = ' INTO ' + ', '.join(into) if into else ''
    return f"EXECUTE format('{sql}'{args}){using}{into};"


def get_update_paths_function_creation(
    model: Type[Model], path_field_lookup: str,
):
    meta = model._meta
    pk = meta.pk
    path_field = meta.get_field(path_field_lookup)
    parent_field = path_field.parent_field
    order_by = path_field.order_by
    if not (pk.attname in order_by or pk.name in order_by
            or 'pk' in order_by):
        order_by = [*order_by, 'pk']

    # TODO: Handle related lookups in `order_by`.
    where_columns = []
    sql_order_by = []
    sql_reversed_order_by = []
    for field_name in order_by:
        descending = field_name[0] == '-'
        if descending:
            field_name = field_name[1:]
        field = (meta.pk if field_name == 'pk'
                 else meta.get_field(field_name))
        quoted_field_name = quote_ident(field.attname)
        where_columns.append(quoted_field_name)
        sql_order_by.append(
            f'{quoted_field_name} {"DESC" if descending else "ASC"}'
        )
        sql_reversed_order_by.append(
            f'{quoted_field_name} {"ASC" if descending else "DESC"}'
        )

    table = meta.db_table
    pk = quote_ident(meta.pk.attname)
    parent = quote_ident(parent_field.attname)
    path = quote_ident(path_field.attname)
    sql_t2_order_by = ', '.join([
        f't2.{ordered_column}' for ordered_column in sql_order_by
    ])
    sql_order_by = ', '.join(sql_order_by)
    sql_reversed_order_by = ', '.join(sql_reversed_order_by)

    rebuild = execute_format(f"""
        -- TODO: Handle concurrent writes during this query (using FOR UPDATE).
        WITH RECURSIVE generate_paths(pk, path) AS ((
                SELECT {parent}, NULL::decimal[]
                FROM {table}
                WHERE {parent} IS NULL
                LIMIT 1
            ) UNION ALL (
                SELECT
                    t2.{pk},
                    t1.path || (
                        row_number() OVER (
                            PARTITION BY t1.pk ORDER BY {sql_t2_order_by}
                        ) - 1
                    )::decimal
                FROM generate_paths AS t1
                INNER JOIN {table} AS t2 ON (
                    t2.{parent} = t1.pk
                    OR (t1.pk IS NULL AND t2.{parent} IS NULL))
            )
        ), updated AS (
            UPDATE {table} AS t2 SET {path} = t1.path
            FROM generate_paths AS t1
            WHERE t2.{pk} = t1.pk AND t2.{pk} != $1.{pk}
                AND (t2.{path} IS NULL OR t2.{path} != t1.path)
        )
        SELECT path FROM generate_paths
        WHERE pk = $1.{pk}
        LIMIT 1
    """, using=['OLD'], into=[f'NEW.{path}'])

    get_new_parent_path = execute_format(f"""
        SELECT {path} FROM {table} WHERE {pk} = $1.{parent}
    """, using=['NEW'], into=['new_parent_path'])

    get_prev_sibling_decimal = execute_format(f"""
        SELECT {path}[array_length({path}, 1)]
        FROM {table}
        WHERE
            {get_prev_sibling_where_clause(where_columns, '$1')}
            AND {compare_columns(parent, f'$1.{parent}')}
            AND {pk} != $1.{pk}
        ORDER BY {sql_reversed_order_by}
        LIMIT 1
    """, using=['NEW'], into=['prev_sibling_decimal'])

    get_next_sibling_decimal = execute_format(f"""
        SELECT {path}[array_length({path}, 1)]
        FROM {table}
        WHERE
            {get_next_sibling_where_clause(where_columns, '$1')}
            AND {compare_columns(parent, f'$1.{parent}')}
            AND {pk} != $1.{pk}
        ORDER BY {sql_order_by}
        LIMIT 1
    """, using=['NEW'], into=['next_sibling_decimal'])

    update_descendants = execute_format(f"""
        UPDATE {table}
        SET {path} = $1.{path}
            || {path}[array_length($2.{path}, 1) + 1:]
        WHERE {path}[:array_length($2.{path}, 1)] = $2.{path} AND {pk} != $2.{pk}
    """, using=['NEW', 'OLD'])

    row_unchanged = join_and([
        compare_columns(f'OLD.{where_column}', f'NEW.{where_column}')
        for where_column in [parent, *where_columns]
    ])

    return f"""
        CREATE OR REPLACE FUNCTION update_{table}_{path}_paths() RETURNS trigger AS $$
        DECLARE
            prev_sibling_decimal decimal := NULL;
            next_sibling_decimal decimal := NULL;
            new_parent_path decimal[];
        BEGIN
            IF TG_OP = 'UPDATE' THEN
                IF NEW.{path} = '{{NULL}}'::decimal[] THEN
                    {rebuild}
                    RETURN NEW;
                END IF;

                IF {row_unchanged} THEN
                    RETURN NEW;
                END IF;
            END IF;

            {get_prev_sibling_decimal}
            {get_next_sibling_decimal}

            IF prev_sibling_decimal IS NULL AND next_sibling_decimal IS NULL THEN
                prev_sibling_decimal := 0;
                next_sibling_decimal := 0;
            ELSE
                IF prev_sibling_decimal IS NULL THEN
                    -- We use `- 2` so that the middle between prev and next
                    -- will be next - 1, that way the new lower bound
                    -- is the former lower bound - 1.
                    prev_sibling_decimal := coalesce(next_sibling_decimal, 0) - 2;
                END IF;
                IF next_sibling_decimal IS NULL THEN
                    -- We use `+ 2` so that the middle between prev and next
                    -- will be prev + 1, that way the new upper bound
                    -- is the former upper bound + 1.
                    next_sibling_decimal := coalesce(prev_sibling_decimal, 0) + 2;
                END IF;
            END IF;
            
            -- Preserve the current path when it is still relevant,
            -- even though it might not be at the middle between prev and next.
            IF TG_OP = 'UPDATE'
                AND {compare_columns(f'NEW.{parent}', f'OLD.{parent}')}
                AND OLD.{path}[array_length(OLD.{path}, 1)]
                    BETWEEN prev_sibling_decimal AND next_sibling_decimal
            THEN
                RETURN NEW;
            END IF;
            
            {get_new_parent_path}
            IF TG_OP = 'UPDATE' THEN
                IF new_parent_path[:array_length(OLD.{path}, 1)] = OLD.{path} THEN
                    RAISE 'Cannot set itself or a descendant as parent.';
                END IF;
            END IF;
            
            NEW.{path} = new_parent_path || (
                prev_sibling_decimal + next_sibling_decimal
            ) / 2;

            IF TG_OP = 'UPDATE' THEN
                {update_descendants}
            END IF;

            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """


CREATE_TRIGGER_QUERIES = (
    """
    CREATE TRIGGER "update_{path}_before"
    BEFORE INSERT OR UPDATE OF {update_columns}
    ON "{table}"
    FOR EACH ROW
    WHEN (pg_trigger_depth() = 0)
    EXECUTE FUNCTION update_{table}_{path}_paths();
    """,
    """
    CREATE OR REPLACE FUNCTION rebuild_{table}_{path}() RETURNS void AS $$
    BEGIN
        UPDATE {table} SET {path} = '{{NULL}}'::decimal[] FROM (
            SELECT * FROM {table}
            WHERE {parent} IS NULL
            LIMIT 1
            FOR UPDATE
        ) AS t
        WHERE {table}.{pk} = t.{pk};
    END;
    $$ LANGUAGE plpgsql;
    """,
    # TODO: Find a way to create this unique constraint
    #       somewhere else.
    """
    ALTER TABLE "{table}"
    ADD CONSTRAINT "{table}_{path}_unique" UNIQUE ("{path}")
    -- FIXME: Remove this `INITIALLY DEFERRED` whenever possible.
    INITIALLY DEFERRED;
    """,
)

DROP_TRIGGER_QUERIES = (
    # TODO: Find a way to delete this unique constraint
    #       somewhere else.
    'ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "{table}_{path}_unique";'
    'DROP TRIGGER IF EXISTS "update_{path}_before" ON "{table}";',
    'DROP FUNCTION IF EXISTS update_{table}_{path}_paths();',
)


def rebuild(table, path_field, db_alias=DEFAULT_DB_ALIAS):
    with connections[db_alias].cursor() as cursor:
        cursor.execute(f'SELECT rebuild_{table}_{path_field}();')


def disable_trigger(table, path_field, db_alias=DEFAULT_DB_ALIAS):
    with connections[db_alias].cursor() as cursor:
        cursor.execute(
            f'ALTER TABLE "{table}" '
            f'DISABLE TRIGGER "update_{path_field}_before";'
        )


def enable_trigger(table, path_field, db_alias=DEFAULT_DB_ALIAS):
    with connections[db_alias].cursor() as cursor:
        cursor.execute(
            f'ALTER TABLE "{table}" '
            f'ENABLE TRIGGER "update_{path_field}_before";'
        )



# coding: utf-8

from __future__ import unicode_literals

import decimal

from django.core.exceptions import ValidationError
from django.db import transaction, InternalError, connection
from django.test import TransactionTestCase

from .models import Place, Person


# TODO: Test same order_by values.
# TODO: Test order_by with descending orders.
# TODO: Test what happens when we move a node after itself
#       while staying in the same siblinghood
#       (it should not create a hole at the former position).
# TODO: Test ORM update/delete.
# TODO: Test raw SQL insertion/update/delete.
# TODO: Test if rebuild works with NULL path values.
# TODO: Test using Path objects as sql parameters.
# TODO: Test multiple path fields on the same model.
# TODO: Test `disable_trigger`, `enable_trigger`, & `disabled_trigger`.
# TODO: Test if `disabled_trigger` does not affect
#       a concurrent node creation/update.
# TODO: Test if breaking a transaction reverts the changes done by the trigger
#       when updating nodes during that transaction.
# TODO: Test non-integer primary keys.
# TODO: Test other `on_delete` behaviour than `CASCADE`.
# TODO: Test unusual table names.


def path(*path_components):
    return [decimal.Decimal(f'{value:.10f}') for value in path_components]


class CommonTest(TransactionTestCase):
    maxDiff = 1000

    def create_place(self, name, parent=None):
        with self.assertNumQueries(1):
            p = Place.objects.create(name=name, parent=parent)
        with self.assertNumQueries(1):
            p.clean()
        # We fetch the object again to populate the path.
        return Place.objects.get(pk=p.pk)

    def create_test_places(self):
        self.correct_raw_places_data = [
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ]
        self.correct_places_data = [
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, 0), 'Eure'),
            (path(0, 0, 1), 'Manche'),
            (path(0, 0, 2), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ]
        france = self.create_place('France')
        yield france
        normandie = self.create_place('Normandie', france)
        yield normandie
        yield self.create_place('Seine-Maritime', normandie)
        yield self.create_place('Eure', normandie)
        yield self.create_place('Manche', normandie)
        osterreich = self.create_place('Österreich')
        yield osterreich
        vienne = self.create_place('Vienne', osterreich)
        yield vienne
        poitou_charentes = self.create_place('Poitou-Charentes', france)
        yield poitou_charentes
        yield self.create_place('Poitiers', vienne)
        vienne.parent = poitou_charentes
        vienne.clean()
        vienne.save()
        yield vienne

    def create_all_test_places(self):
        list(self.create_test_places())

    def assertPlaces(self, values, queryset=None, n_queries=1):
        with self.assertNumQueries(n_queries):
            if queryset is None:
                queryset = Place.objects.all()
            places = list(queryset)
            self.assertListEqual([(p.path.value, p.name) for p in places],
                                 values)


class PathTest(CommonTest):
    maxDiff = None

    def test_path_on_creation(self):
        with self.assertNumQueries(1):
            place1 = Place.objects.create(name='place1')
        # 1 query because the path got deferred,  forcing Django to run
        # a new query to get the updated value when we need it. Same below.
        with self.assertNumQueries(1):
            self.assertEqual(place1.path.value, path(0))
        with self.assertNumQueries(1):
            place2 = Place.objects.create(name='place2', parent=place1)
        with self.assertNumQueries(1):
            self.assertEqual(place2.path.value, path(0, 0))
        with self.assertNumQueries(1):
            place2.parent = None
            place2.clean()
            place2.save()
        with self.assertNumQueries(1):
            self.assertEqual(place2.path.value, path(1))

    def test_insert(self):
        it = self.create_test_places()
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, 0), 'Seine-Maritime'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, 0), 'Seine-Maritime'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(1), 'Österreich'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(1), 'Österreich'),
            (path(1, 0), 'Vienne'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(1), 'Österreich'),
            (path(1, 0), 'Vienne'),
        ])
        next(it)
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(1), 'Österreich'),
            (path(1, 0), 'Vienne'),
            (path(1, 0, 0), 'Poitiers'),
        ])
        next(it)
        self.assertPlaces(self.correct_raw_places_data)
        Place.rebuild_paths()
        self.assertPlaces(self.correct_places_data)

    def test_delete(self):
        self.create_all_test_places()

        self.assertPlaces(self.correct_raw_places_data)

        # Leaf
        manche = Place.objects.get(name='Manche')
        with self.assertNumQueries(3):
            manche.delete()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

        # Branch
        normandie = Place.objects.get(name='Normandie')
        with self.assertNumQueries(3):
            normandie.delete()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

        # Root
        france = Place.objects.get(name='France')
        with self.assertNumQueries(3):
            france.delete()
        self.assertPlaces([
            (path(1), 'Österreich'),
        ])

    def test_move_root_to_prev_root(self):
        self.create_all_test_places()

        osterreich = Place.objects.get(name='Österreich')
        osterreich.name = 'Autriche'
        with self.assertNumQueries(1):
            osterreich.clean()
            osterreich.save()
        self.assertPlaces([
            (path(-1), 'Autriche'),
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'Autriche'),
            (path(1), 'France'),
            (path(1, 0), 'Normandie'),
            (path(1, 0, 0), 'Eure'),
            (path(1, 0, 1), 'Manche'),
            (path(1, 0, 2), 'Seine-Maritime'),
            (path(1, 1), 'Poitou-Charentes'),
            (path(1, 1, 0), 'Vienne'),
            (path(1, 1, 0, 0), 'Poitiers'),
        ])

    def test_move_root_to_next_root(self):
        self.create_all_test_places()

        france = Place.objects.get(name='France')
        france.name = 'République française'
        with self.assertNumQueries(1):
            france.clean()
            france.save()
        self.assertPlaces([
            (path(1), 'Österreich'),
            (path(2), 'République française'),
            (path(2, 0), 'Normandie'),
            (path(2, 0, -1), 'Eure'),
            (path(2, 0, -0.5), 'Manche'),
            (path(2, 0, 0), 'Seine-Maritime'),
            (path(2, 1), 'Poitou-Charentes'),
            (path(2, 1, 0), 'Vienne'),
            (path(2, 1, 0, 0), 'Poitiers'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'Österreich'),
            (path(1), 'République française'),
            (path(1, 0), 'Normandie'),
            (path(1, 0, 0), 'Eure'),
            (path(1, 0, 1), 'Manche'),
            (path(1, 0, 2), 'Seine-Maritime'),
            (path(1, 1), 'Poitou-Charentes'),
            (path(1, 1, 0), 'Vienne'),
            (path(1, 1, 0, 0), 'Poitiers'),
        ])

    def test_move_root_to_prev_branch(self):
        self.create_all_test_places()

        little_france = Place.objects.create(name='Île-de-France')
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(0.5), 'Île-de-France'),
            (path(1), 'Österreich'),
        ])

        little_france.parent = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            little_france.clean()
        with self.assertNumQueries(1):
            little_france.save()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, -1), 'Île-de-France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Île-de-France'),
            (path(0, 1), 'Normandie'),
            (path(0, 1, 0), 'Eure'),
            (path(0, 1, 1), 'Manche'),
            (path(0, 1, 2), 'Seine-Maritime'),
            (path(0, 2), 'Poitou-Charentes'),
            (path(0, 2, 0), 'Vienne'),
            (path(0, 2, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

    def test_move_root_to_next_branch(self):
        self.create_all_test_places()

        bretagne = Place.objects.create(name='Bretagne')
        self.assertPlaces([
            (path(-1), 'Bretagne'),
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

        bretagne.parent = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            bretagne.clean()
        with self.assertNumQueries(1):
            bretagne.save()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, -1), 'Bretagne'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Bretagne'),
            (path(0, 1), 'Normandie'),
            (path(0, 1, 0), 'Eure'),
            (path(0, 1, 1), 'Manche'),
            (path(0, 1, 2), 'Seine-Maritime'),
            (path(0, 2), 'Poitou-Charentes'),
            (path(0, 2, 0), 'Vienne'),
            (path(0, 2, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

    def test_move_root_to_prev_leaf(self):
        self.create_all_test_places()

        grattenoix = Place.objects.create(name='Grattenoix')
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(0.5), 'Grattenoix'),
            (path(1), 'Österreich'),
        ])

        grattenoix.parent = Place.objects.get(name='Seine-Maritime')
        with self.assertNumQueries(1):
            grattenoix.clean()
        with self.assertNumQueries(1):
            grattenoix.save()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 0, 0, 0), 'Grattenoix'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, 0), 'Eure'),
            (path(0, 0, 1), 'Manche'),
            (path(0, 0, 2), 'Seine-Maritime'),
            (path(0, 0, 2, 0), 'Grattenoix'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

    def test_move_root_to_next_leaf(self):
        self.create_all_test_places()

        evreux = Place.objects.create(name='Évreux')
        self.assertPlaces([
            (path(-1), 'Évreux'),
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

        evreux.parent = Place.objects.get(name='Eure')
        with self.assertNumQueries(1):
            evreux.clean()
        with self.assertNumQueries(1):
            evreux.save()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -1, 0), 'Évreux'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces([
            (path(0), 'France'),
            (path(0, 0), 'Normandie'),
            (path(0, 0, 0), 'Eure'),
            (path(0, 0, 0, 0), 'Évreux'),
            (path(0, 0, 1), 'Manche'),
            (path(0, 0, 2), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'),
            (path(1), 'Österreich'),
        ])

    # TODO: Add move_branch_to_prev_root.
    # TODO: Add move_branch_to_next_root.
    # TODO: Add move_branch_to_prev_branch.
    # TODO: Add move_branch_to_next_branch.
    # TODO: Add move_branch_to_prev_leaf.
    # TODO: Add move_branch_to_next_leaf.
    # TODO: Add move_leaf_to_prev_root.
    # TODO: Add move_leaf_to_next_root.
    # TODO: Add move_leaf_to_prev_branch.
    # TODO: Add move_leaf_to_next_branch.
    # TODO: Add move_leaf_to_prev_leaf.
    # TODO: Add move_leaf_to_next_leaf.

    def test_get_level(self):
        self.create_all_test_places()

        with self.assertNumQueries(1):
            data = [(p.get_level(), p.name) for p in Place.objects.all()]
            self.assertListEqual(data, [
                (1, 'France'),
                (2, 'Normandie'),
                (3, 'Eure'),
                (3, 'Manche'),
                (3, 'Seine-Maritime'),
                (2, 'Poitou-Charentes'),
                (3, 'Vienne'),
                (4, 'Poitiers'),
                (1, 'Österreich'),
            ])

    def test_is_root(self):
        self.create_all_test_places()

        places = [p.name for p in Place.objects.all() if p.is_root()]
        self.assertListEqual(places, ['France', 'Österreich'])

    def test_is_leaf(self):
        self.create_all_test_places()

        places = [p.name for p in Place.objects.all() if p.is_leaf()]
        self.assertListEqual(places, ['Eure', 'Manche', 'Seine-Maritime',
                                      'Poitiers', 'Österreich'])

    def test_get_children(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_children()
                     .values_list('name', flat=True)),
                ['Normandie', 'Poitou-Charentes'])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_children()
                     .values_list('name', flat=True)),
                ['Eure', 'Manche', 'Seine-Maritime'])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_children()
                     .values_list('name', flat=True)), [])

    def test_get_ancestors(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_ancestors(include_self=True)
                     .values_list('name', flat=True)), ['France'])

        with self.assertNumQueries(0):
            self.assertListEqual(
                list(france.get_ancestors()
                     .values_list('name', flat=True)), [])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_ancestors(include_self=True)
                     .values_list('name', flat=True)), ['France', 'Normandie'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_ancestors()
                     .values_list('name', flat=True)), ['France'])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_ancestors(include_self=True)
                     .values_list('name', flat=True)),
                ['France', 'Normandie', 'Seine-Maritime'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_ancestors()
                     .values_list('name', flat=True)), ['France', 'Normandie'])

    def test_get_descendants(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_descendants(include_self=True)
                     .values_list('name', flat=True)),
                ['France', 'Normandie', 'Eure', 'Manche', 'Seine-Maritime',
                 'Poitou-Charentes', 'Vienne', 'Poitiers'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_descendants()
                     .values_list('name', flat=True)),
                ['Normandie', 'Eure', 'Manche', 'Seine-Maritime',
                 'Poitou-Charentes', 'Vienne', 'Poitiers'])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_descendants(include_self=True)
                     .values_list('name', flat=True)),
                ['Normandie', 'Eure', 'Manche', 'Seine-Maritime'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_descendants()
                     .values_list('name', flat=True)),
                ['Eure', 'Manche', 'Seine-Maritime'])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_descendants(include_self=True)
                     .values_list('name', flat=True)), ['Seine-Maritime'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_descendants()
                     .values_list('name', flat=True)), [])

    def test_get_siblings(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['France', 'Österreich'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_siblings()
                     .values_list('name', flat=True)), ['Österreich'])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['Normandie', 'Poitou-Charentes'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_siblings()
                     .values_list('name', flat=True)), ['Poitou-Charentes'])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['Eure', 'Manche', 'Seine-Maritime'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_siblings()
                     .values_list('name', flat=True)), ['Eure', 'Manche'])

    def test_filtered_get_siblings(self):
        self.create_all_test_places()
        queryset = Place.objects.filter(name__lt='O')

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_siblings(include_self=True, queryset=queryset)
                     .values_list('name', flat=True)),
                ['France'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_siblings(queryset=queryset)
                     .values_list('name', flat=True)), [])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_siblings(include_self=True,
                                            queryset=queryset)
                     .values_list('name', flat=True)),
                ['Normandie'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_siblings(queryset=queryset)
                     .values_list('name', flat=True)), [])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_siblings(include_self=True,
                                                 queryset=queryset)
                     .values_list('name', flat=True)),
                ['Eure', 'Manche'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_siblings(queryset=queryset)
                     .values_list('name', flat=True)), ['Eure', 'Manche'])

    def test_get_prev_siblings(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_prev_siblings(include_self=True)
                     .values_list('name', flat=True)), ['France'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_prev_siblings()
                     .values_list('name', flat=True)), [])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_prev_siblings(include_self=True)
                     .values_list('name', flat=True)), ['Normandie'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_prev_siblings()
                     .values_list('name', flat=True)), [])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_prev_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['Seine-Maritime', 'Manche', 'Eure'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_prev_siblings()
                     .values_list('name', flat=True)), ['Manche', 'Eure'])

    def test_get_next_siblings(self):
        self.create_all_test_places()

        # Root

        france = Place.objects.get(name='France')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_next_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['France', 'Österreich'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(france.get_next_siblings()
                     .values_list('name', flat=True)), ['Österreich'])

        # Branch

        normandie = Place.objects.get(name='Normandie')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_next_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['Normandie', 'Poitou-Charentes'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(normandie.get_next_siblings()
                     .values_list('name', flat=True)), ['Poitou-Charentes'])

        # Leaf

        seine_maritime = Place.objects.get(name='Seine-Maritime')

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_next_siblings(include_self=True)
                     .values_list('name', flat=True)),
                ['Seine-Maritime'])

        with self.assertNumQueries(1):
            self.assertListEqual(
                list(seine_maritime.get_next_siblings()
                     .values_list('name', flat=True)), [])

    def test_get_prev_sibling(self):
        self.create_all_test_places()

        # Root
        france = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            self.assertIsNone(france.get_prev_sibling())

        # Branch
        normandie = Place.objects.get(name='Normandie')
        with self.assertNumQueries(1):
            self.assertIsNone(normandie.get_prev_sibling())

        # Leaf
        seine_maritime = Place.objects.get(name='Seine-Maritime')
        with self.assertNumQueries(1):
            self.assertEqual(
                seine_maritime.get_prev_sibling().name, 'Manche')

    def test_get_next_sibling(self):
        self.create_all_test_places()

        # Root
        france = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            self.assertEqual(france.get_next_sibling().name, 'Österreich')

        # Branch
        normandie = Place.objects.get(name='Normandie')
        with self.assertNumQueries(1):
            self.assertEqual(
                normandie.get_next_sibling().name, 'Poitou-Charentes')

        # Leaf
        seine_maritime = Place.objects.get(name='Seine-Maritime')
        with self.assertNumQueries(1):
            self.assertIsNone(seine_maritime.get_next_sibling())

    def test_filtered_get_prev_sibling(self):
        self.create_all_test_places()
        queryset = Place.objects.filter(name__lt='O')

        # Root
        france = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            self.assertIsNone(france.get_prev_sibling(queryset=queryset))

        # Branch
        normandie = Place.objects.get(name='Normandie')
        with self.assertNumQueries(1):
            self.assertIsNone(normandie.get_prev_sibling(queryset=queryset))

        # Leaf
        seine_maritime = Place.objects.get(name='Seine-Maritime')
        with self.assertNumQueries(1):
            self.assertEqual(
                seine_maritime.get_prev_sibling(queryset=queryset).name,
                'Manche')

    def test_filtered_get_next_sibling(self):
        self.create_all_test_places()
        queryset = Place.objects.filter(name__lt='P')

        # Root
        france = Place.objects.get(name='France')
        with self.assertNumQueries(1):
            self.assertEqual(france.get_next_sibling(queryset=queryset).name,
                             'Österreich')

        # Branch
        normandie = Place.objects.get(name='Normandie')
        with self.assertNumQueries(1):
            self.assertIsNone(normandie.get_next_sibling(queryset=queryset))

        # Leaf
        seine_maritime = Place.objects.get(name='Seine-Maritime')
        with self.assertNumQueries(1):
            self.assertIsNone(
                seine_maritime.get_next_sibling(queryset=queryset))

    def test_new_path(self):
        place = Place()

        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_children()), [])
        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_ancestors()), [])
        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_descendants()), [])
        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_siblings()), [])
        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_prev_siblings()), [])
        with self.assertNumQueries(0):
            self.assertListEqual(list(place.get_next_siblings()), [])
        with self.assertNumQueries(0):
            self.assertIsNone(place.get_prev_sibling())
        with self.assertNumQueries(0):
            self.assertIsNone(place.get_next_sibling())
        with self.assertNumQueries(0):
            self.assertIsNone(place.get_level())
        with self.assertNumQueries(0):
            self.assertIsNone(place.is_root())
        with self.assertNumQueries(0):
            self.assertIsNone(place.is_leaf())

    def test_comparisons(self):
        self.create_all_test_places()

        france = Place.objects.get(name='France').path
        self.assertTrue(france == france)
        self.assertFalse(france != france)
        self.assertFalse(france > france)
        self.assertTrue(france >= france)
        self.assertFalse(france < france)
        self.assertTrue(france <= france)

        # vs None
        self.assertFalse(france == '')
        self.assertTrue(france != '')
        self.assertTrue(france < '')
        self.assertTrue(france <= '')
        self.assertFalse(france > '')
        self.assertFalse(france >= '')

        # vs None
        self.assertFalse(france == None)
        self.assertTrue(france != None)
        self.assertTrue(france < None)
        self.assertTrue(france <= None)
        self.assertFalse(france > None)
        self.assertFalse(france >= None)

        # vs new node
        new_node = Place().path
        self.assertFalse(france == new_node)
        self.assertTrue(france != new_node)
        self.assertTrue(france < new_node)
        self.assertTrue(france <= new_node)
        self.assertFalse(france > new_node)
        self.assertFalse(france >= new_node)

        # Same level
        osterreich = Place.objects.get(name='Österreich').path
        self.assertEqual(france.get_level(), osterreich.get_level())
        self.assertFalse(france == osterreich)
        self.assertTrue(france != osterreich)
        self.assertTrue(france < osterreich)
        self.assertTrue(france <= osterreich)
        self.assertFalse(france > osterreich)
        self.assertFalse(france >= osterreich)

        # Inferior level
        normandie = Place.objects.get(name='Normandie').path
        self.assertLess(france.get_level(), normandie.get_level())
        self.assertFalse(france == normandie)
        self.assertTrue(france != normandie)
        self.assertTrue(france < normandie)
        self.assertTrue(france <= normandie)
        self.assertFalse(france > normandie)
        self.assertFalse(france >= normandie)

        # Superior level
        self.assertGreater(normandie.get_level(), osterreich.get_level())
        self.assertFalse(normandie == osterreich)
        self.assertTrue(normandie != osterreich)
        self.assertTrue(normandie < osterreich)
        self.assertTrue(normandie <= osterreich)
        self.assertFalse(normandie > osterreich)
        self.assertFalse(normandie >= osterreich)

    def test_is_ancestor_of(self):
        self.create_all_test_places()

        for place in Place.objects.all():
            self.assertFalse(place.is_ancestor_of(place))
            self.assertTrue(place.is_ancestor_of(place,
                                                 include_self=True))
            for ancestor in place.get_ancestors():
                self.assertTrue(ancestor.is_ancestor_of(place))

    def test_is_descendant_of(self):
        self.create_all_test_places()

        for place in Place.objects.all():
            self.assertFalse(place.is_descendant_of(place))
            self.assertTrue(place.is_descendant_of(place,
                                                   include_self=True))
            for descendant in place.get_descendants():
                self.assertTrue(descendant.is_descendant_of(place))

    def test_filter_roots(self):
        self.create_all_test_places()

        self.assertPlaces([
            (path(0), 'France'),
            (path(1), 'Österreich'),
        ], queryset=Place.objects.filter_roots())

    def test_rebuild(self):
        self.create_all_test_places()

        with Place.disabled_tree_trigger():
            updated_places = []
            for i, place in enumerate(Place.objects.order_by('name')):
                place.path = [i]
                updated_places.append(place)
            Place.objects.bulk_update(updated_places, ['path'])
        self.assertPlaces([
            (path(0), 'Eure'), (path(1), 'France'), (path(2), 'Manche'),
            (path(3), 'Normandie'), (path(4), 'Österreich'), (path(5), 'Poitiers'),
            (path(6), 'Poitou-Charentes'), (path(7), 'Seine-Maritime'),
            (path(8), 'Vienne')])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces(self.correct_places_data)

        # Root
        with Place.disabled_tree_trigger():
            Place.objects.filter(name='France').update(path=[89])
        self.assertPlaces([
            (path(0, 0), 'Normandie'), (path(0, 0, 0), 'Eure'),
            (path(0, 0, 1), 'Manche'), (path(0, 0, 2), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'), (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'), (path(1), 'Österreich'),
            (path(89), 'France')])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces(self.correct_places_data)

        # Branch
        with Place.disabled_tree_trigger():
            Place.objects.filter(name='Normandie').update(path=[89, 89])
        self.assertPlaces([
            (path(0), 'France'), (path(0, 0, 0), 'Eure'),
            (path(0, 0, 1), 'Manche'), (path(0, 0, 2), 'Seine-Maritime'),
            (path(0, 1), 'Poitou-Charentes'), (path(0, 1, 0), 'Vienne'),
            (path(0, 1, 0, 0), 'Poitiers'), (path(1), 'Österreich'),
            (path(89, 89), 'Normandie')])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces(self.correct_places_data)

        # Leaf
        with Place.disabled_tree_trigger():
            Place.objects.filter(name='Seine-Maritime').update(path=[0, 89])
        self.assertPlaces([
            (path(0), 'France'), (path(0, 0), 'Normandie'), (path(0, 0, 0), 'Eure'),
            (path(0, 0, 1), 'Manche'), (path(0, 1), 'Poitou-Charentes'),
            (path(0, 1, 0), 'Vienne'), (path(0, 1, 0, 0), 'Poitiers'),
            (path(0, 89), 'Seine-Maritime'), (path(1), 'Österreich')])
        with self.assertNumQueries(1):
            Place.rebuild_paths()
        self.assertPlaces(self.correct_places_data)

    def test_cycle(self):
        # Simple cycle
        a = Place.objects.create(name='a')
        a.parent = a

        with self.assertRaisesMessage(
            ValidationError,
            "{'parent': [\"Value 'a' is not a valid choice.\"]}",
        ):
            with transaction.atomic():
                with self.assertNumQueries(1):
                    a.clean()

        with self.assertRaisesMessage(
            InternalError, 'Cannot set itself or a descendant as parent.',
        ):
            with transaction.atomic():
                with self.assertNumQueries(1):
                    a.save()

        # Complex cycle
        b = Place.objects.create(name='b', parent=a)
        c = Place.objects.create(name='c', parent=b)
        d = Place.objects.create(name='d', parent=c)
        a.parent = d

        with self.assertRaisesMessage(
            ValidationError,
            "{'parent': [\"Value 'd' is not a valid choice.\"]}",
        ):
            with transaction.atomic():
                with self.assertNumQueries(1):
                    a.clean()

        with self.assertRaisesMessage(
                InternalError, 'Cannot set itself or a descendant as parent.'):
            with transaction.atomic():
                with self.assertNumQueries(1):
                    a.save()

    def test_path_in_cursor(self):
        place1 = self.create_place('place1')
        with connection.cursor() as cursor:
            cursor.execute('SELECT %s;', [place1.path])
        place2 = self.create_place('place2', place1)
        with connection.cursor() as cursor:
            cursor.execute('SELECT %s;', [place2.path])


class MultipleOrderByFieldsTest(TransactionTestCase):
    maxDiff = None

    def setUp(self):
        self.correct_raw_persons_data = [
            (path(-3), 18, 'Leopold', 'Mozart'),
            (path(-3, -1), 18, 'Maria Anna', 'Mozart'),
            (path(-3, 0), 18, 'Wolfgang Amadeus', 'Mozart'),
            (path(-1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(-0.75), 19, 'Johann (father)', 'Strauss'),
            (path(-0.75, 0), 19, 'Johann (son)', 'Strauss'),
            (path(-0.5), 19, 'Piotr Ilyich', 'Tchaikovski'),
            (path(-0.25), 20, '', 'Strauss'),
            (path(0), None, '', 'Anonymous'),
        ]
        self.correct_persons_data = [
            (path(0), 18, 'Leopold', 'Mozart'),
            (path(0, 0), 18, 'Maria Anna', 'Mozart'),
            (path(0, 1), 18, 'Wolfgang Amadeus', 'Mozart'),
            (path(1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(2), 19, 'Johann (father)', 'Strauss'),
            (path(2, 0), 19, 'Johann (son)', 'Strauss'),
            (path(3), 19, 'Piotr Ilyich', 'Tchaikovski'),
            (path(4), 20, '', 'Strauss'),
            (path(5), None, '', 'Anonymous'),
        ]
        self.anonymous = Person.objects.create(
            last_name='Anonymous',
        )
        self.vivaldi = Person.objects.create(
            century=18, first_name='Antonio Lucio', last_name='Vivaldi',
        )
        self.wolfgang_mozart = Person.objects.create(
            century=18, first_name='Wolfgang Amadeus', last_name='Mozart',
        )
        self.leopold_mozart = Person.objects.create(
            century=18, first_name='Leopold', last_name='Mozart',
        )
        self.wolfgang_mozart.parent = self.leopold_mozart
        self.wolfgang_mozart.clean()
        self.wolfgang_mozart.save()
        self.maria_anna_mozart = Person.objects.create(
            parent=self.leopold_mozart,
            century=18, first_name='Maria Anna', last_name='Mozart',
        )
        self.tchaikovski = Person.objects.create(
            century=19, first_name='Piotr Ilyich', last_name='Tchaikovski',
        )
        self.strauss_father = Person.objects.create(
            century=19, first_name='Johann (father)', last_name='Strauss',
        )
        self.strauss_son = Person.objects.create(
            parent=self.strauss_father,
            century=19, first_name='Johann (son)', last_name='Strauss',
        )
        self.strauss = Person.objects.create(
            century=20, last_name='Strauss',
        )

    def assertPersons(self, values, queryset=None, n_queries=1):
        with self.assertNumQueries(n_queries):
            if queryset is None:
                queryset = Person.objects.all()
            persons = list(queryset)
            self.assertListEqual(
                [
                    (p.path.value, p.century, p.first_name, p.last_name)
                    for p in persons
                ],
                values,
            )

    def test_rebuild(self):
        self.assertPersons(self.correct_raw_persons_data)
        with Person.disabled_tree_trigger():
            for i, person in enumerate(
                Person.objects.order_by('-last_name', '-first_name')
            ):
                # We add 10 to make sure
                # we do not clash with the existing paths.
                person.path = [10 + i]
                person.clean()
                person.save()
        self.assertPersons([
            (path(10), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(11), 19, 'Piotr Ilyich', 'Tchaikovski'),
            (path(12), 19, 'Johann (son)', 'Strauss'),
            (path(13), 19, 'Johann (father)', 'Strauss'),
            (path(14), 20, '', 'Strauss'),
            (path(15), 18, 'Wolfgang Amadeus', 'Mozart'),
            (path(16), 18, 'Maria Anna', 'Mozart'),
            (path(17), 18, 'Leopold', 'Mozart'),
            (path(18), None, '', 'Anonymous'),
        ])
        Person.rebuild_paths()
        self.assertPersons(self.correct_persons_data)

    def test_clash_on_insert(self):
        """
        Checks that instances with exactly the same `order_by` values
        are assigned different paths on insertion, sorted by primary key.
        """
        vivaldi2 = Person.objects.create(
            century=18, first_name='Antonio Lucio', last_name='Vivaldi',
        )
        self.assertGreater(vivaldi2.pk, self.vivaldi.pk)
        self.assertGreater(vivaldi2.path, self.vivaldi.path)
        self.assertPersons([
            (path(-1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(-0.875), 18, 'Antonio Lucio', 'Vivaldi'),
        ], queryset=Person.objects.filter(last_name='Vivaldi'))
        Person.rebuild_paths()
        self.assertPersons([
            (path(1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(2), 18, 'Antonio Lucio', 'Vivaldi'),
        ], queryset=Person.objects.filter(last_name='Vivaldi'))

    def test_clash_on_update(self):
        """
        Checks that instances with exactly the same `order_by` values
        are assigned different paths on update, sorted by primary key.
        """
        vivaldi2 = Person.objects.create(
            century=18, first_name='Some', last_name='Guy',
        )
        self.assertGreater(vivaldi2.pk, self.vivaldi.pk)
        self.assertLess(vivaldi2.path, self.vivaldi.path)
        self.assertPersons([
            (path(-4), 18, 'Some', 'Guy'),
            (path(-1), 18, 'Antonio Lucio', 'Vivaldi'),
        ], queryset=Person.objects.filter(last_name__in=['Vivaldi', 'Guy']))
        vivaldi2.first_name = 'Antonio Lucio'
        vivaldi2.last_name = 'Vivaldi'
        vivaldi2.clean()
        vivaldi2.save()
        self.assertPersons([
            (path(-1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(-0.875), 18, 'Antonio Lucio', 'Vivaldi'),
        ], queryset=Person.objects.filter(last_name='Vivaldi'))
        Person.rebuild_paths()
        self.assertPersons([
            (path(1), 18, 'Antonio Lucio', 'Vivaldi'),
            (path(2), 18, 'Antonio Lucio', 'Vivaldi'),
        ], queryset=Person.objects.filter(last_name='Vivaldi'))


class QuerySetTest(CommonTest):
    def test_get_descendants(self):
        self.create_all_test_places()

        places = Place.objects.filter(name__in=('Normandie', 'Österreich'))
        self.assertPlaces([
            (path(0, 0), 'Normandie'),
            (path(1), 'Österreich'),
        ], places)

        self.assertPlaces([
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
        ], places.get_descendants())
        self.assertPlaces([
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(1), 'Österreich'),
        ], places.get_descendants(include_self=True))

        osterreich = Place.objects.get(name='Österreich')
        self.create_place('Vienne (AU)', osterreich)

        self.assertPlaces([
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(1, 0), 'Vienne (AU)'),
        ], places.get_descendants())
        self.assertPlaces([
            (path(0, 0), 'Normandie'),
            (path(0, 0, -1), 'Eure'),
            (path(0, 0, -0.5), 'Manche'),
            (path(0, 0, 0), 'Seine-Maritime'),
            (path(1), 'Österreich'),
            (path(1, 0), 'Vienne (AU)'),
        ], places.get_descendants(include_self=True))



from contextlib import contextmanager

from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections, transaction
from django.db.models import DecimalField, Index, Func, F
from django.db.models.expressions import RawSQL
from django.utils.translation import ugettext_lazy as _

from .sql import postgresql
from .types import Path


# TODO: Handle ManyToManyField('self') instead of ForeignKey('self').
# TODO: Add queryset methods like `get_descendants` in a mixin.
# TODO: Implement an alternative for other db backends.


class PathField(ArrayField):
    description = _('Tree path')

    @classmethod
    def get_indexes(
        cls,
        table_name: str,
        path_field_name: str,
        max_indexed_level: int = 5,
    ):
        return [
            Index(
                # TODO: Simplify using `trim_array`
                #       once support for PostgreSQL < 14 is dropped.
                RawSQL(
                    f'{path_field_name}[:array_length({path_field_name}, 1) - 1]',
                    ()
                ),
                name=f'{table_name}_{path_field_name}_parent_index',
            ),
            Index(
                F(f'{path_field_name}__level'),
                name=f'{table_name}_{path_field_name}_level_index',
            ),
            *[
                Index(
                    F(f'{path_field_name}__0_{level}'),
                    name=f'{table_name}_{path_field_name}_slice_{level}_index',
                )
                for level in range(1, max_indexed_level + 1)
            ]
        ]

    def __init__(self, *args, parent_field_name: str = 'parent', **kwargs):
        for kwarg in ('base_field', 'default', 'null', 'unique'):
            if kwarg in kwargs:
                raise ImproperlyConfigured('Cannot set `PathField.%s`.'
                                           % kwarg)

        kwargs['base_field'] = DecimalField(max_digits=20, decimal_places=10)
        kwargs['default'] = lambda: Path(self, None)
        kwargs.setdefault('editable', False)
        kwargs['null'] = True

        self.order_by = list(kwargs.pop('order_by', []))
        self.parent_field_name = parent_field_name

        super(PathField, self).__init__(*args, **kwargs)

    @property
    def parent_field(self):
        return self.model._meta.get_field(self.parent_field_name)

    def contribute_to_class(self, cls, name, *args, **kwargs):
        if name in self.order_by:
            raise ImproperlyConfigured(
                '`PathField.order_by` cannot reference itself.' % name)
        super(PathField, self).contribute_to_class(cls, name, *args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(PathField, self).deconstruct()
        del kwargs['base_field']
        if not kwargs['editable']:
            del kwargs['editable']
        del kwargs['default']
        del kwargs['null']
        del kwargs['size']
        if self.order_by:
            kwargs['order_by'] = self.order_by
        if self.parent_field_name != 'parent':
            kwargs['parent_field_name'] = self.parent_field_name
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        if isinstance(value, Path):
            return value
        return Path(self, value)

    def to_python(self, value):
        if isinstance(value, Path):
            return value
        return Path(self, value)

    def get_prep_value(self, value):
        if isinstance(value, Path):
            return value.value
        return value

    def _check_database_backend(self, db_alias):
        if connections[db_alias].vendor != 'postgresql':
            raise NotImplementedError(
                'django-tree is only for PostgreSQL for now.')

    def rebuild(self, db_alias=DEFAULT_DB_ALIAS):
        self._check_database_backend(db_alias)
        postgresql.rebuild(self.model._meta.db_table, self.attname,
                           db_alias=db_alias)

    def disable_trigger(self, db_alias=DEFAULT_DB_ALIAS):
        self._check_database_backend(db_alias)
        postgresql.disable_trigger(self.model._meta.db_table, self.attname,
                                   db_alias=db_alias)

    def enable_trigger(self, db_alias=DEFAULT_DB_ALIAS):
        self._check_database_backend(db_alias)
        postgresql.enable_trigger(self.model._meta.db_table, self.attname,
                                  db_alias=db_alias)

    @contextmanager
    @transaction.atomic
    def disabled_trigger(self, db_alias=DEFAULT_DB_ALIAS):
        self.disable_trigger(db_alias=db_alias)
        try:
            yield
        finally:
            self.enable_trigger(db_alias=db_alias)



from contextlib import contextmanager
from typing import Optional

from django.core.exceptions import ValidationError
from django.db import DEFAULT_DB_ALIAS, transaction
from django.db.models import Model

from .query import _get_path_fields, _get_path_field, TreeManager


class TreeModelMixin:
    @classmethod
    def _get_path_fields(cls, name: Optional[str] = None):
        return _get_path_fields(cls, name)

    @classmethod
    def _get_path_field(cls, name):
        return _get_path_field(cls, name)

    def _get_path_value(self, path_field):
        return getattr(self, self._get_path_field(path_field).name)

    def get_children(self, path_field=None):
        return self._get_path_value(path_field).get_children()

    def get_ancestors(self, include_self=False, path_field=None):
        return (self._get_path_value(path_field)
                .get_ancestors(include_self=include_self))

    def get_descendants(self, include_self=False, path_field=None):
        return (self._get_path_value(path_field)
                .get_descendants(include_self=include_self))

    def get_siblings(self, include_self=False, queryset=None, path_field=None):
        return (self._get_path_value(path_field)
                .get_siblings(include_self=include_self, queryset=queryset))

    def get_prev_siblings(self, include_self=False, queryset=None,
                          path_field=None):
        return (self._get_path_value(path_field)
                .get_prev_siblings(include_self=include_self,
                                   queryset=queryset))

    def get_next_siblings(self, include_self=False, queryset=None,
                          path_field=None):
        return (self._get_path_value(path_field)
                .get_next_siblings(include_self=include_self,
                                   queryset=queryset))

    def get_prev_sibling(self, queryset=None, path_field=None):
        return (self._get_path_value(path_field)
                .get_prev_sibling(queryset=queryset))

    def get_next_sibling(self, queryset=None, path_field=None):
        return (self._get_path_value(path_field)
                .get_next_sibling(queryset=queryset))

    def get_level(self, path_field=None):
        return self._get_path_value(path_field).get_level()

    def is_root(self, path_field=None):
        return self._get_path_value(path_field).is_root()

    def is_leaf(self, path_field=None):
        return self._get_path_value(path_field).is_leaf()

    def is_ancestor_of(self, other, include_self=False, path_field=None):
        return (self._get_path_value(path_field)
                .is_ancestor_of(other._get_path_value(path_field),
                                include_self=include_self))

    def is_descendant_of(self, other, include_self=False, path_field=None):
        return (self._get_path_value(path_field)
                .is_descendant_of(other._get_path_value(path_field),
                                  include_self=include_self))

    def clean(self):
        super().clean()
        if not self._state.adding:
            for path_field in self._get_path_fields():
                old_path = getattr(self, path_field.attname)
                parent_field = path_field.parent_field
                new_parent = getattr(self, parent_field.name)
                if not new_parent:
                    continue

                if not isinstance(new_parent, Model):
                    try:
                        new_parent = self.__class__._default_manager.get(
                            pk=new_parent,
                        )
                    except self.__class__.DoesNotExist:
                        new_parent = self

                new_parent_path = getattr(new_parent, path_field.attname)
                if new_parent_path.is_descendant_of(
                    old_path, include_self=True,
                ):
                    raise ValidationError({
                        parent_field.name: ValidationError(
                            parent_field.error_messages['invalid_choice'],
                            code='invalid_choice',
                            params={'value': str(new_parent)},
                        )
                    })

    def delete(self, using=None, **kwargs):
        assert self.pk is not None, (
            "%s object can't be deleted because "
            "its %s attribute is set to None." %
            (self._meta.object_name, self._meta.pk.attname)
        )
        qs = self.get_descendants(include_self=True)
        if using is not None:
            qs = qs.using(using)
        return qs.delete()

    @classmethod
    def rebuild_paths(cls, db_alias=DEFAULT_DB_ALIAS, path_field=None):
        """
        Rebuilds the paths of all the ``PathField``s
        if ``path_field`` is ``None``.  Otherwise, only paths from
        the ``PathField`` with the ``path_field`` name are rebuilt.
        """

        for field in cls._get_path_fields(path_field):
            field.rebuild(db_alias=db_alias)

    @classmethod
    def disable_tree_trigger(cls, db_alias=DEFAULT_DB_ALIAS, path_field=None):
        for field in cls._get_path_fields(path_field):
            field.disable_trigger(db_alias=db_alias)

    @classmethod
    def enable_tree_trigger(cls, db_alias=DEFAULT_DB_ALIAS, path_field=None):
        for field in cls._get_path_fields(path_field):
            field.enable_trigger(db_alias=db_alias)

    @classmethod
    @contextmanager
    @transaction.atomic
    def disabled_tree_trigger(cls, db_alias=DEFAULT_DB_ALIAS, path_field=None):
        """
        Context manager for temporarily disabling django-tree triggers.

        If ``path_field`` is ``None``, disables all the triggers.
        Otherwise disables only the trigger
        """

        cls.disable_tree_trigger(db_alias=db_alias, path_field=path_field)
        try:
            yield
        finally:
            cls.enable_tree_trigger(db_alias=db_alias, path_field=path_field)


class TreeModel(TreeModelMixin, Model):
    objects = TreeManager()

    class Meta:
        abstract = True



from django.db import migrations
from django.db.models import (
    AutoField, CharField, ForeignKey, PositiveIntegerField, Manager, CASCADE,
    Index, F,
)
from django.db.models.expressions import RawSQL
from mptt.fields import TreeForeignKey
from tree.fields import PathField
from tree.operations import CreateTreeTrigger

from ..models import get_random_name


class Migration(migrations.Migration):
    dependencies = [
        ('tree', '0001_initial'),
    ]

    operations = [

        migrations.CreateModel(
            name='MPTTPlace',
            fields=[
                ('id', AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', CharField(max_length=50, unique=True, default=get_random_name)),
                ('lft', PositiveIntegerField(db_index=True, editable=False)),
                ('rght', PositiveIntegerField(db_index=True, editable=False)),
                ('tree_id', PositiveIntegerField(db_index=True, editable=False)),
                ('level', PositiveIntegerField(db_index=True, editable=False)),
                ('parent', TreeForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
            ],
            managers=[
                ('objects', Manager()),
            ],
        ),

        migrations.CreateModel(
            name='TreePlace',
            fields=[
                ('id', AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', CharField(max_length=50, unique=True, default=get_random_name)),
                ('parent', ForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
                ('path', PathField(order_by=['name'], db_index=True)),
            ],
            options={
                'indexes': [
                    Index(RawSQL('path[:array_length(path, 1) - 1]', ()), name='treeplace_path_parent_index'),
                    Index(F('path__level'), name='treeplace_path_level_index'),
                    Index(F('path__0_1'), name='treeplace_path_slice_1_index'),
                    Index(F('path__0_2'), name='treeplace_path_slice_2_index'),
                    Index(F('path__0_3'), name='treeplace_path_slice_3_index'),
                    Index(F('path__0_4'), name='treeplace_path_slice_4_index'),
                    Index(F('path__0_5'), name='treeplace_path_slice_5_index'),
                ],
            },
        ),
        CreateTreeTrigger('TreePlace'),

        migrations.CreateModel(
            name='TreebeardALPlace',
            fields=[
                ('id', AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', CharField(max_length=50, unique=True, default=get_random_name)),
                ('parent', ForeignKey('self', blank=True, null=True, on_delete=CASCADE)),
            ],
        ),

        migrations.CreateModel(
            name='TreebeardMPPlace',
            fields=[
                ('id', AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', CharField(max_length=255, unique=True)),
                ('depth', PositiveIntegerField()),
                ('numchild', PositiveIntegerField(default=0)),
                ('name', CharField(max_length=50, unique=True, default=get_random_name)),
            ],
        ),

        migrations.CreateModel(
            name='TreebeardNSPlace',
            fields=[
                ('id', AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('lft', PositiveIntegerField(db_index=True)),
                ('rgt', PositiveIntegerField(db_index=True)),
                ('tree_id', PositiveIntegerField(db_index=True)),
                ('depth', PositiveIntegerField(db_index=True)),
                ('name', CharField(max_length=50, unique=True, default=get_random_name)),
            ],
        ),

    ]



from django.apps import AppConfig

from .fields import PathField
from .lookups import AncestorOf, SiblingOf, ChildOf, DescendantOf
from .transforms import Level


class TreeAppConfig(AppConfig):
    name = 'tree'
    verbose_name = 'Tree'

    def ready(self):
        PathField.register_lookup(AncestorOf)
        PathField.register_lookup(SiblingOf)
        PathField.register_lookup(ChildOf)
        PathField.register_lookup(DescendantOf)

        PathField.register_lookup(Level)

        # Loads signals
        from . import signals



from django.db.models import QuerySet


class Path:
    def __init__(self, field, value):
        self.field = field
        self.attname = getattr(self.field, 'attname', None)
        self.field_bound = self.attname is not None
        self.qs = (self.field.model._default_manager.all()
                   if self.field_bound else QuerySet())
        self.value = value

    def __repr__(self):
        if self.field_bound:
            return '<Path %s %s>' % (self.field, self.value)
        return '<Path %s>' % self.value

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, Path):
            other = other.value
        return self.value == other

    def __ne__(self, other):
        if isinstance(other, Path):
            other = other.value
        return self.value != other

    def __lt__(self, other):
        # We simulate the effects of a NULLS LAST.
        if not self.value:
            return False
        if isinstance(other, Path):
            other = other.value
        if not other:
            return True
        return self.value < other

    def __le__(self, other):
        # We simulate the effects of a NULLS LAST.
        if not self.value:
            return False
        if isinstance(other, Path):
            other = other.value
        if not other:
            return True
        return self.value <= other

    def __gt__(self, other):
        # We simulate the effects of a NULLS LAST.
        if not self.value:
            return True
        if isinstance(other, Path):
            other = other.value
        if not other:
            return False
        return self.value > other

    def __ge__(self, other):
        # We simulate the effects of a NULLS LAST.
        if not self.value:
            return True
        if isinstance(other, Path):
            other = other.value
        if not other:
            return False
        return self.value >= other

    def __iter__(self):
        return iter(self.value)

    def get_children(self):
        if not self.value:
            return self.qs.none()
        return self.qs.filter(**{
            f'{self.attname}__child_of': self.value,
        })

    def get_ancestors(self, include_self=False):
        if not self.value or (self.is_root() and not include_self):
            return self.qs.none()
        path = self.value
        if not include_self:
            path = path[:-1]
        # Using the lookup `ancestor_of` here is slower,
        # so we explicitly specify the ancestors’ paths.
        return self.qs.filter(**{self.attname + '__in': [
            path[:i] for i in range(1, len(path) + 1)
        ]})

    def get_descendants(self, include_self=False):
        if not self.value:
            return self.qs.none()
        # Using the lookup `descendant_of` here is slower,
        # so we explicitly specify the path slicing comparison.
        qs = self.qs.filter(**{
            f'{self.attname}__0_{len(self.value)}': self.value,
        })
        if include_self:
            return qs
        return qs.filter(**{f'{self.attname}__level__gt': len(self.value)})

    def get_siblings(self, include_self=False, queryset=None):
        if not self.value:
            return self.qs.none()

        qs = self.qs if queryset is None else queryset
        qs = qs.filter(**{
            self.attname + '__sibling_of': self.value,
        })
        if include_self:
            return qs
        return qs.exclude(**{self.attname: self.value})

    def get_prev_siblings(self, include_self=False, queryset=None):
        if not self.value:
            return self.qs.none()
        siblings = self.get_siblings(include_self=include_self,
                                     queryset=queryset)
        lookup = '__lte' if include_self else '__lt'
        return (siblings.filter(**{self.attname + lookup: self.value})
                .order_by('-' + self.attname))

    def get_next_siblings(self, include_self=False, queryset=None):
        if not self.value:
            return self.qs.none()
        siblings = self.get_siblings(include_self=include_self,
                                     queryset=queryset)
        lookup = '__gte' if include_self else '__gt'
        return (siblings.filter(**{self.attname + lookup: self.value})
                .order_by(self.attname))

    def get_prev_sibling(self, queryset=None):
        if not self.value:
            return

        return self.get_prev_siblings(queryset=queryset).first()

    def get_next_sibling(self, queryset=None):
        if not self.value:
            return None

        return self.get_next_siblings(queryset=queryset).first()

    def get_level(self):
        if self.value:
            return len(self.value)

    def is_root(self):
        if self.value:
            return len(self.value) == 1

    def is_leaf(self):
        if self.value:
            return not self.get_children().exists()

    def is_ancestor_of(self, other, include_self=False):
        if not self.value:
            return False
        if isinstance(other, Path):
            other = other.value
        if not other:
            return False
        if not isinstance(other, list):
            raise TypeError(
                '`other` must be a `Path` instance or a list of decimals.'
            )
        if not include_self and self.value == other:
            return False
        return other[:len(self.value)] == self.value

    def is_descendant_of(self, other, include_self=False):
        if not self.value:
            return False
        if isinstance(other, Path):
            other = other.value
        if not other:
            return False
        if not isinstance(other, list):
            raise TypeError(
                '`other` must be a `Path` instance or a list of decimals.'
            )
        if not include_self and self.value == other:
            return False
        return self.value[:len(other)] == other


# Tells psycopg2 how to prepare a Path object for the database,
# in case it doesn't go through the ORM.
try:
    import psycopg2
except ImportError:
    pass
else:
    from psycopg2.extensions import register_adapter, adapt

    def adapt_path(path):
        return adapt(path.value)

    register_adapter(Path, adapt_path)



from django.db.migrations.operations.base import Operation

from .sql import postgresql
from .sql.base import quote_ident


class CheckDatabaseMixin:
    def check_database_backend(self, schema_editor):
        if schema_editor.connection.vendor != 'postgresql':
            raise NotImplementedError(
                'django-tree is only for PostgreSQL for now.')


class GetModelMixin:
    def get_model(self, app_label, state):
        get_model = state.apps.get_model
        return (get_model(self.model_lookup) if '.' in self.model_lookup
                else get_model(app_label, self.model_lookup))


class CreateTreeTrigger(Operation, GetModelMixin, CheckDatabaseMixin):
    reversible = True
    atomic = True

    def __init__(self, model_lookup, path_field='path'):
        self.model_lookup = model_lookup
        self.path_field_lookup = path_field

    def get_pre_params(self, model):
        meta = model._meta
        path_field = meta.get_field(self.path_field_lookup)
        parent_field = path_field.parent_field
        order_by = path_field.order_by

        # TODO: Handle related lookups in `order_by`.
        path = quote_ident(path_field.attname)
        update_columns = [path]
        for field_name in order_by:
            descending = field_name[0] == '-'
            if descending:
                field_name = field_name[1:]

            if field_name == 'pk':
                continue

            quoted_field_name = quote_ident(meta.get_field(field_name).attname)
            update_columns.append(quoted_field_name)

        return dict(
            table=quote_ident(meta.db_table),
            pk=quote_ident(meta.pk.attname),
            parent=quote_ident(parent_field.attname),
            path=path,
            update_columns=', '.join(update_columns),
        )

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor,
                          from_state, to_state):
        self.check_database_backend(schema_editor)
        model = self.get_model(app_label, to_state)
        # We escape the modulo operator '%' otherwise Django considers it
        # as a placeholder for a parameter.
        schema_editor.execute(
            postgresql.get_update_paths_function_creation(
                model=model,
                path_field_lookup=self.path_field_lookup,
            ).replace('%', '%%')
        )
        for sql_query in postgresql.CREATE_TRIGGER_QUERIES:
            schema_editor.execute(sql_query.format(
                **self.get_pre_params(model=model)))

    def database_backwards(self, app_label, schema_editor,
                           from_state, to_state):
        self.check_database_backend(schema_editor)
        for sql_query in postgresql.DROP_TRIGGER_QUERIES:
            schema_editor.execute(sql_query.format(
                **self.get_pre_params(self.get_model(app_label, to_state))))

    def describe(self):
        return 'Creates a trigger that automatically updates a `PathField`'


class DeleteTreeTrigger(CreateTreeTrigger):
    def database_forwards(self, *args, **kwargs):
        super(DeleteTreeTrigger, self).database_backwards(*args, **kwargs)

    def database_backwards(self, *args, **kwargs):
        super(DeleteTreeTrigger, self).database_forwards(*args, **kwargs)

    def describe(self):
        return 'Deletes the trigger that automatically updates a `PathField`'


class RebuildPaths(Operation, GetModelMixin, CheckDatabaseMixin):
    reversible = True
    atomic = True

    def __init__(self, model_lookup, path_field='path'):
        self.model_lookup = model_lookup
        self.path_field = path_field
        super(RebuildPaths, self).__init__()

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor,
                          from_state, to_state):
        self.check_database_backend(schema_editor)
        model = self.get_model(app_label, to_state)
        postgresql.rebuild(model._meta.db_table, self.path_field,
                           db_alias=schema_editor.connection.alias)

    def database_backwards(self, app_label, schema_editor,
                           from_state, to_state):
        self.check_database_backend(schema_editor)

    def describe(self):
        return 'Rebuilds all the tree structure of a given django-tree field'



from __future__ import print_function
from collections import Iterable
import os
from time import time
from typing import Type, List, Optional

from django.db import connections, router, transaction
from django.db.models import Max, F, Model
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
import pandas as pd
from tqdm import tqdm

from .models import (
    TreePlace, MPTTPlace, TreebeardMPPlace, TreebeardNSPlace, TreebeardALPlace,
)
from .utils import prefix_unit, SkipTest


DISK_USAGE = 'Disk usage (bytes)'
READ_LATENCY = 'Read latency (s)'
WRITE_LATENCY = 'Write latency (s)'

BYTES_FORMATTER = FuncFormatter(lambda v, pos: prefix_unit(v, 'B', -3))
SECONDS_FORMATTER = FuncFormatter(lambda v, pos: prefix_unit(v, 's'))


class Benchmark:
    models = {
        MPTTPlace: 'MPTT',
        TreePlace: 'tree',
        TreebeardALPlace: 'treebeard AL',
        TreebeardMPPlace: 'treebeard MP',
        TreebeardNSPlace: 'treebeard NS',
    }
    siblings_per_level = (
        5, 5, 5, 5, 5,
    )
    tests = {}
    ticks_formatters = {
        DISK_USAGE: BYTES_FORMATTER,
        READ_LATENCY: SECONDS_FORMATTER,
        WRITE_LATENCY: SECONDS_FORMATTER,
    }
    results_path = 'benchmark/results/'

    def __init__(
        self,
        run_django_tree_only: bool = False,
        db_optimization_interval: int = 100,
        selected_tests: Optional[List[str]] = None,
    ):
        self.run_django_tree_only = run_django_tree_only
        self.db_optimization_interval = db_optimization_interval
        self.selected_tests = selected_tests
        self.data = []
        self.router = router.routers[0]

        self.rows_count = 0
        n = 1
        for i in self.siblings_per_level:
            n *= i
            self.rows_count += n

        if not os.path.exists(self.results_path):
            os.makedirs(self.results_path)

    @property
    def current_db_alias(self):
        return self.router.db_alias

    @current_db_alias.setter
    def current_db_alias(self, db_alias):
        self.router.db_alias = db_alias

    def add_data(self, model, test_name, count, value, y_label=READ_LATENCY):
        self.data.append({
            'Database': connections[self.current_db_alias].vendor,
            'Test name': test_name,
            'Count': count,
            'Implementation': self.models[model],
            'Value': value,
            'Y label': y_label,
        })

    def populate_database(self, model, level=1, parents=(None,)):
        n_siblings = self.siblings_per_level[level-1]
        for parent in parents:
            if model in (TreePlace, TreebeardALPlace):
                bulk = [model(parent=parent)
                        for _ in range(n_siblings)]
                model.objects.bulk_create(bulk)
                objects = model.objects.filter(parent=parent)
            elif model in (TreebeardMPPlace, TreebeardNSPlace):
                # We fetch again each parent because the path can change
                # during the creation of children from the previous parent.
                if parent is not None:
                    parent = model.objects.get(pk=parent.pk)
                objects = [model.add_root() if parent is None
                           else parent.add_child() for _ in range(n_siblings)]
            else:
                objects = [model.objects.create(parent=parent)
                           for _ in range(n_siblings)]
            yield model.objects.count()
            if level < len(self.siblings_per_level):
                for count in self.populate_database(model, level+1, objects):
                    yield count

    def create_databases(self):
        old_db_names = {}
        for alias in connections:
            conn = connections[alias]
            old_db_names[alias] = conn.settings_dict['NAME']
            conn.creation.create_test_db(autoclobber=True)

    @classmethod
    def register_test(cls, name, models=None, y_label=READ_LATENCY):
        if models is None:
            models = cls.models
        if not isinstance(models, Iterable):
            models = (models,)

        def inner(test_class):
            for model in models:
                cls.tests[(name, model, y_label)] = test_class
            return test_class

        return inner

    def skip_test(self, test_name: str) -> bool:
        return self.selected_tests and test_name not in self.selected_tests

    def run_tests(self, tested_model, count):
        connection = connections[self.current_db_alias]
        for (test_name, model, y_label), test_class in self.tests.items():
            if model is not tested_model or self.skip_test(test_name):
                continue
            benchmark_test = test_class(self, model)
            try:
                benchmark_test.setup()
            except SkipTest:
                value = elapsed_time = None
            else:
                start = time()
                if benchmark_test.rollback:
                    with transaction.atomic(using=self.current_db_alias):
                        value = benchmark_test.run()
                        connection.needs_rollback = True
                else:
                    value = benchmark_test.run()
                elapsed_time = time() - start
            if value is None:
                value = elapsed_time
            self.add_data(model, test_name, count, value, y_label=y_label)

    def plot(self, df, database_name, test_name, y_label):
        means = df.rolling(max(df.index.max() // 20, 1)).mean()
        ax = means.plot(
            title=test_name, alpha=0.8,
            xlim=(0, means.index.max() * 1.05),
            ylim=(0, means.max().max() * 1.05),
        )
        ax.set(xlabel='Amount of objects in table', ylabel=y_label)

        ax.xaxis.set_major_formatter(
            FuncFormatter(lambda v, pos: prefix_unit(v, '', -3)))
        if y_label in self.ticks_formatters:
            ax.yaxis.set_major_formatter(self.ticks_formatters[y_label])

        legend = ax.legend(
            loc='upper center', bbox_to_anchor=(0.5, 0.0),
            bbox_transform=plt.gcf().transFigure,
            fancybox=True, shadow=True, ncol=3)

        filename = ('%s - %s.svg' % (database_name,
                                     test_name)).replace(' ', '_')
        plt.savefig(
            os.path.join(self.results_path, filename),
            bbox_extra_artists=(legend,), bbox_inches='tight',
        )

    def force_update_db_stats_and_indexes(self, model: Type[Model]):
        with connections[self.current_db_alias].cursor() as cursor:
            # This makes sure the table statistics are
            # up to date and the disk usage is optimised.
            cursor.execute(
                'VACUUM FULL ANALYZE "%s";' % model._meta.db_table)
            # This makes sure the indexes are up to date.
            cursor.execute(
                'REINDEX TABLE "%s";' % model._meta.db_table)

    def run(self):
        self.create_databases()

        for db_alias in connections:
            self.current_db_alias = db_alias
            connection = connections[db_alias]

            for model in sorted(self.models, key=lambda m: m.__name__):
                if self.run_django_tree_only and model is not TreePlace:
                    continue
                print('-' * 50)
                print('%s on %s' % (self.models[model], connection.vendor))
                it = self.populate_database(model)
                progress = tqdm(it, total=self.rows_count)
                elapsed_time = 0.0
                while True:
                    try:
                        start = time() - elapsed_time
                        count = next(it)
                        elapsed_time = time() - start
                    except StopIteration:
                        break
                    progress.update(count - progress.n)
                    self.add_data(model, 'Create all objects', count,
                                  elapsed_time, y_label=WRITE_LATENCY)
                    if count % self.db_optimization_interval == 0:
                        self.force_update_db_stats_and_indexes(model)
                    self.run_tests(model, count)
                # We delete the objects to avoid impacting
                # the following tests and to clear some disk space.
                model.objects.all().delete()

        csv_path = os.path.join(self.results_path, 'data.csv')
        if self.run_django_tree_only:
            df = pd.read_csv(csv_path)
            df = df[df['Implementation'] != self.models[TreePlace]]
            df = df.append(pd.DataFrame(self.data))
        else:
            df = pd.DataFrame(self.data)
        df.to_csv(csv_path, index=False)

        stats_df = df.set_index(['Database', 'Test name', 'Count'])
        stats_df.sort_index(inplace=True)
        group_by = stats_df.groupby(level=[0, 1, 2])
        min_values_series = group_by.min()['Value']
        max_values_series = group_by.max()['Value']
        stats_df['Value'] = ((stats_df['Value'] - min_values_series)
                             / (max_values_series - min_values_series))
        stats_df['Value'] = stats_df['Value'].fillna(0.0)
        stats_df = stats_df.groupby(['Y label', 'Implementation']).mean()
        stats_df['Value'] = ((20 * (1 - stats_df['Value']))
                             .apply(lambda f: '%.1f / 20' % f))
        stats_df.to_html(os.path.join(self.results_path, 'stats.html'),
                         header=False)

        df.set_index('Count', inplace=True)
        for database_name in df['Database'].unique():
            for test_name in df['Test name'].unique():
                if self.skip_test(test_name):
                    continue
                sub_df = df[(df['Database'] == database_name)
                            & (df['Test name'] == test_name)]
                y_labels = sub_df['Y label'].unique()
                assert len(y_labels) == 1
                sub_df = sub_df.pivot(columns='Implementation', values='Value')
                self.plot(sub_df, database_name, test_name, y_labels[0])


class BenchmarkTest:
    rollback: bool = False

    def __init__(self, benchmark, model):
        self.benchmark = benchmark
        self.model = model

    def setup(self):
        pass

    def run(self):
        raise NotImplementedError


class BenchmarkWriteTest(BenchmarkTest):
    rollback = True


@Benchmark.register_test('Table disk usage (including indexes)',
                         y_label=DISK_USAGE)
class TestDiskUsage(BenchmarkTest):
    def run(self):
        with connections[self.benchmark.current_db_alias].cursor() as cursor:
            cursor.execute("SELECT pg_total_relation_size('%s');"
                           % self.model._meta.db_table)
            return cursor.fetchone()[0]


class GetRootMixin:
    def setup(self):
        qs = self.model._default_manager.all()
        qs = (qs.filter(depth=1)
              if self.model in (TreebeardMPPlace, TreebeardNSPlace)
              else qs.filter(parent__isnull=True))
        self.root = qs[qs.count() // 2]
        super().setup()


class GetBranchMixin:
    def setup(self):
        super().setup()

        qs = self.model._default_manager.all()
        if hasattr(self, 'root'):
            descendants = self.root.get_descendants()
            if isinstance(descendants, list):
                descendants = [d.pk for d in descendants]
            qs = qs.exclude(pk__in=descendants)

        if self.model is MPTTPlace:
            qs = qs.filter(level=1)
        elif self.model is TreePlace:
            qs = qs.filter(path__len=2)
        elif self.model is TreebeardALPlace:
            qs = qs.filter(parent__isnull=False, parent__parent__isnull=True)
        else:
            qs = qs.filter(depth=2)
        try:
            self.branch = qs[qs.count() // 2]
        except IndexError:
            raise SkipTest


class GetLeafMixin:
    def setup(self):
        super(GetLeafMixin, self).setup()

        qs = self.model._default_manager.all()
        if hasattr(self, 'root'):
            descendants = self.root.get_descendants()
            if isinstance(descendants, list):
                descendants = [d.pk for d in descendants]
            qs = qs.exclude(pk=self.root.pk).exclude(pk__in=descendants)
        if hasattr(self, 'branch'):
            descendants = self.branch.get_descendants()
            if isinstance(descendants, list):
                descendants = [d.pk for d in descendants]
            qs = qs.exclude(pk=self.branch.pk).exclude(pk__in=descendants)

        qs = (qs.annotate(n=Max('depth')).filter(depth=F('n'), depth__gt=1)
              if self.model in (TreebeardMPPlace, TreebeardNSPlace)
              else qs.filter(children__isnull=True, parent__isnull=False))
        try:
            self.leaf = qs[qs.count() // 2]
        except IndexError:
            raise SkipTest


#
# Children
#


@Benchmark.register_test('Get children [root]')
class TestGetChildrenRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        list(self.root.get_children())


@Benchmark.register_test('Get children [branch]')
class TestGetChildrenBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        list(self.branch.get_children())


@Benchmark.register_test('Get children [leaf]')
class TestGetChildrenLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        list(self.leaf.get_children())


#
# Children count
#


@Benchmark.register_test('Get children count [root]',
                         (MPTTPlace, TreePlace))
class TestGetChildrenCountRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_children().count()


@Benchmark.register_test(
    'Get children count [root]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetChildrenCountRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_children_count()


@Benchmark.register_test('Get children count [branch]',
                         (MPTTPlace, TreePlace))
class TestGetChildrenCountBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_children().count()


@Benchmark.register_test(
    'Get children count [branch]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetChildrenCountBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_children_count()


@Benchmark.register_test('Get children count [leaf]',
                         (MPTTPlace, TreePlace))
class TestGetChildrenCountLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_children().count()


@Benchmark.register_test(
    'Get children count [leaf]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetChildrenCounteaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_children_count()


#
# Filtered children count
#


@Benchmark.register_test('Get filtered children count [root]')
class TestGetFilteredChildrenCountRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_children().filter(pk__contains='1').count()


@Benchmark.register_test('Get filtered children count [branch]')
class TestGetFilteredChildrenCountBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_children().filter(pk__contains='1').count()


@Benchmark.register_test('Get filtered children count [leaf]')
class TestGetFilteredChildrenCountLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_children().filter(pk__contains='1').count()


#
# Ancestors
#


@Benchmark.register_test('Get ancestors [root]')
class TestGetAncestorsRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        list(self.root.get_ancestors())


@Benchmark.register_test('Get ancestors [branch]',)
class TestGetAncestorsBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        list(self.branch.get_ancestors())


@Benchmark.register_test('Get ancestors [leaf]')
class TestGetAncestorsLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        list(self.leaf.get_ancestors())


#
# Descendants
#


@Benchmark.register_test('Get descendants [root]')
class TestGetDescendantsRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        list(self.root.get_descendants())


@Benchmark.register_test('Get descendants [branch]')
class TestGetDescendantsBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        list(self.branch.get_descendants())


@Benchmark.register_test('Get descendants [leaf]')
class TestGetDescendantsLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        list(self.leaf.get_descendants())


@Benchmark.register_test('Get descendants from queryset',
                         (MPTTPlace, TreePlace))
class TestGetDescendantsFromQuerySet(BenchmarkTest):
    def setup(self):
        self.qs = self.model._default_manager.annotate(
            n=F('pk') % 5).filter(n=0)
        super().setup()

    def run(self):
        list(self.qs.get_descendants())


#
# Descendants count
#


@Benchmark.register_test('Get descendants count [root]', TreePlace)
class TestGetDescendantsCountRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_descendants().count()


@Benchmark.register_test(
    'Get descendants count [root]',
    (MPTTPlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetDescendantsCountRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_descendant_count()


@Benchmark.register_test('Get descendants count [branch]', TreePlace)
class TestGetDescendantsCountBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_descendants().count()


@Benchmark.register_test(
    'Get descendants count [branch]',
    (MPTTPlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetDescendantsCountBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_descendant_count()


@Benchmark.register_test('Get descendants count [leaf]', TreePlace)
class TestGetDescendantsCountLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_descendants().count()


@Benchmark.register_test(
    'Get descendants count [leaf]',
    (MPTTPlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetDescendantsCountLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_descendant_count()


#
# Filtered descendants count
#


@Benchmark.register_test('Get filtered descendants count [root]')
class TestGetFilteredDescendantsCountRoot(GetRootMixin, BenchmarkTest):
    def setup(self):
        if self.model is TreebeardALPlace:
            raise SkipTest
        super().setup()

    def run(self):
        self.root.get_descendants().filter(pk__contains='1').count()


@Benchmark.register_test('Get filtered descendants count [branch]')
class TestGetFilteredDescendantsCountBranch(GetBranchMixin, BenchmarkTest):
    def setup(self):
        if self.model is TreebeardALPlace:
            raise SkipTest
        super().setup()

    def run(self):
        self.branch.get_descendants().filter(pk__contains='1').count()


@Benchmark.register_test('Get filtered descendants count [leaf]')
class TestGetFilteredDescendantsCountLeaf(GetLeafMixin, BenchmarkTest):
    def setup(self):
        if self.model is TreebeardALPlace:
            raise SkipTest
        super().setup()

    def run(self):
        self.leaf.get_descendants().filter(pk__contains='1').count()


#
# Siblings
#


@Benchmark.register_test('Get siblings [root]', (MPTTPlace, TreePlace))
class TestGetSiblingsRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        list(self.root.get_siblings(include_self=True))


@Benchmark.register_test(
    'Get siblings [root]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetSiblingsRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        list(self.root.get_siblings())


@Benchmark.register_test('Get siblings [branch]', (MPTTPlace, TreePlace))
class TestGetSiblingsBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        list(self.branch.get_siblings(include_self=True))


@Benchmark.register_test(
    'Get siblings [branch]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetSiblingsBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        list(self.branch.get_siblings())


@Benchmark.register_test('Get siblings [leaf]', (MPTTPlace, TreePlace))
class TestGetSiblingsLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        list(self.leaf.get_siblings(include_self=True))


@Benchmark.register_test(
    'Get siblings [leaf]',
    (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetSiblingsLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        list(self.leaf.get_siblings())


#
# Previous sibling
#


@Benchmark.register_test('Get previous sibling [root]', MPTTPlace)
class TestGetPrevSiblingRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_previous_sibling()


@Benchmark.register_test(
    'Get previous sibling [root]',
    (TreePlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetPrevSiblingRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_prev_sibling()


@Benchmark.register_test('Get previous sibling [branch]', MPTTPlace)
class TestGetPrevSiblingBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_previous_sibling()


@Benchmark.register_test(
    'Get previous sibling [branch]',
    (TreePlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetPrevSiblingBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_prev_sibling()


@Benchmark.register_test('Get previous sibling [leaf]', MPTTPlace)
class TestGetPrevSiblingLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_previous_sibling()


@Benchmark.register_test(
    'Get previous sibling [leaf]',
    (TreePlace, TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetPrevSiblingLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_prev_sibling()


#
# Next sibling
#


@Benchmark.register_test('Get next sibling [root]')
class TestGetNextSiblingRoot(GetRootMixin, BenchmarkTest):
    def run(self):
        self.root.get_next_sibling()


@Benchmark.register_test('Get next sibling [branch]')
class TestGetNextSiblingBranch(GetBranchMixin, BenchmarkTest):
    def run(self):
        self.branch.get_next_sibling()


@Benchmark.register_test('Get next sibling [leaf]')
class TestGetNextSiblingLeaf(GetLeafMixin, BenchmarkTest):
    def run(self):
        self.leaf.get_next_sibling()


#
# Get roots
#


@Benchmark.register_test('Get roots', MPTTPlace)
class TestGetRoots(BenchmarkTest):
    def run(self):
        list(self.model._default_manager.root_nodes())


@Benchmark.register_test('Get roots', TreePlace)
class TestGetRoots(BenchmarkTest):
    def run(self):
        list(self.model.objects.filter_roots())


@Benchmark.register_test(
    'Get roots', (TreebeardALPlace, TreebeardMPPlace, TreebeardNSPlace))
class TestGetRoots(BenchmarkTest):
    def run(self):
        list(self.model.get_root_nodes())


#
# Rebuild
#


@Benchmark.register_test('Rebuild paths', MPTTPlace, y_label=WRITE_LATENCY)
class TestRebuildPaths(BenchmarkWriteTest):
    def run(self):
        self.model._default_manager.rebuild()


@Benchmark.register_test('Rebuild paths', TreePlace, y_label=WRITE_LATENCY)
class TestRebuildPaths(BenchmarkWriteTest):
    def run(self):
        self.model.rebuild_paths()


@Benchmark.register_test('Rebuild paths', (TreebeardALPlace, TreebeardNSPlace),
                         y_label=WRITE_LATENCY)
class TestRebuildPaths(BenchmarkWriteTest):
    def setup(self):
        raise SkipTest


@Benchmark.register_test('Rebuild paths', TreebeardMPPlace,
                         y_label=WRITE_LATENCY)
class TestRebuildPaths(BenchmarkWriteTest):
    def run(self):
        self.model.fix_tree()


#
# Create
#


@Benchmark.register_test(
    'Create [root]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestCreateRoot(BenchmarkWriteTest):
    def run(self):
        self.model.objects.create()


@Benchmark.register_test(
    'Create [root]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestCreateRoot(BenchmarkWriteTest):
    def run(self):
        self.model.add_root()


@Benchmark.register_test(
    'Create [branch]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestCreateBranch(GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.model.objects.create(parent=self.root)


@Benchmark.register_test(
    'Create [branch]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestCreateBranch(GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.add_child()


@Benchmark.register_test(
    'Create [leaf]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestCreateLeaf(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.model.objects.create(parent=self.leaf)


@Benchmark.register_test(
    'Create [leaf]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestCreateLeaf(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.add_child()


#
# Save without any change
#


@Benchmark.register_test('Save without change [root]', y_label=WRITE_LATENCY)
class TestSaveRootWithoutChange(GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.save()


@Benchmark.register_test('Save without change [branch]', y_label=WRITE_LATENCY)
class TestSaveBranchWithoutChange(GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.save()


@Benchmark.register_test('Save without change [leaf]', y_label=WRITE_LATENCY)
class TestSaveLeafWithoutChange(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.save()


#
# Move
#


@Benchmark.register_test('Move [same root path]', y_label=WRITE_LATENCY)
class TestMoveSameRootPath(GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.name += ' 2'
        self.root.save()


@Benchmark.register_test('Move [same branch path]', y_label=WRITE_LATENCY)
class TestMoveSameBranchPath(GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.name += ' 2'
        self.branch.save()


@Benchmark.register_test('Move [same leaf path]', y_label=WRITE_LATENCY)
class TestMoveSameLeafPath(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.name += ' 2'
        self.leaf.save()


@Benchmark.register_test(
    'Move [root to branch]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveRootToBranch(GetBranchMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.parent = self.branch
        self.root.save()


@Benchmark.register_test(
    'Move [root to branch]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveRootToBranch(GetBranchMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.move(self.branch, pos='sorted-child')


@Benchmark.register_test(
    'Move [root to leaf]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveRootToLeaf(GetLeafMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.parent = self.leaf
        self.root.save()


@Benchmark.register_test(
    'Move [root to leaf]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveRootToLeaf(GetLeafMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.move(self.leaf, pos='sorted-child')


@Benchmark.register_test(
    'Move [branch to root]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveBranchToRoot(GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.parent = None
        self.branch.save()


@Benchmark.register_test(
    'Move [branch to root]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveBranchToRoot(GetBranchMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.move(self.root, pos='sorted-sibling')


@Benchmark.register_test(
    'Move [branch to leaf]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveBranchToLeaf(GetLeafMixin, GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.parent = self.leaf
        self.branch.save()


@Benchmark.register_test(
    'Move [branch to leaf]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveBranchToLeaf(GetLeafMixin, GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.move(self.leaf, pos='sorted-child')


@Benchmark.register_test(
    'Move [leaf to root]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveLeafToRoot(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.parent = None
        self.leaf.save()


@Benchmark.register_test(
    'Move [leaf to root]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveLeafToRoot(GetLeafMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.move(self.root, pos='sorted-sibling')


@Benchmark.register_test(
    'Move [leaf to branch]',
    (MPTTPlace, TreePlace, TreebeardALPlace), y_label=WRITE_LATENCY)
class TestMoveLeafToBranch(GetLeafMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.parent = self.root
        self.leaf.save()


@Benchmark.register_test(
    'Move [leaf to branch]',
    (TreebeardMPPlace, TreebeardNSPlace), y_label=WRITE_LATENCY)
class TestMoveLeafToBranch(GetLeafMixin, GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.move(self.root, pos='sorted-child')


#
# Delete
#


@Benchmark.register_test('Delete [root]', y_label=WRITE_LATENCY)
class TestDeleteRoot(GetRootMixin, BenchmarkWriteTest):
    def run(self):
        self.root.delete()


@Benchmark.register_test('Delete [branch]', y_label=WRITE_LATENCY)
class TestDeleteBranch(GetBranchMixin, BenchmarkWriteTest):
    def run(self):
        self.branch.delete()


@Benchmark.register_test('Delete [leaf]', y_label=WRITE_LATENCY)
class TestDeleteLeaf(GetLeafMixin, BenchmarkWriteTest):
    def run(self):
        self.leaf.delete()



from django.db.models import IntegerField, Transform


class Level(Transform):
    lookup_name = 'level'
    template = 'array_length(%(expressions)s, 1)'

    @property
    def output_field(self):
        return IntegerField()



from django.db import DEFAULT_DB_ALIAS


class AutoRouter:
    def __init__(self):
        self.db_alias = DEFAULT_DB_ALIAS

    def db_for_read(self, model, **hints):
        return self.db_alias

    def db_for_write(self, model, **hints):
        return self.db_alias

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        return True



from django.db import migrations


class Migration(migrations.Migration):
    operations = [
        # There used to be an operation here, which is no longer relevant.
        # We keep the migration to avoid messing with existing deployments
        # of django-tree.
    ]



from django.forms import ModelChoiceField


class TreeChoiceField(ModelChoiceField):
    def label_from_instance(self, obj):
        if obj.is_root():
            return str(obj)
        return '%s %s' % ('──' * (obj.get_level()-1), obj)



from django.db.models import CharField, ForeignKey, CASCADE, SmallIntegerField

from tree.fields import PathField
from tree.models import TreeModel


class Place(TreeModel):
    name = CharField(max_length=50)
    parent = ForeignKey('self', null=True, blank=True, on_delete=CASCADE)
    path = PathField(order_by=['name'])

    class Meta:
        ordering = ['path', 'name']
        indexes = [
            *PathField.get_indexes('place', 'path'),
        ]

    def __str__(self):
        return self.name


class Person(TreeModel):
    century = SmallIntegerField(null=True, blank=True)
    first_name = CharField(max_length=20, blank=True)
    last_name = CharField(max_length=50)
    parent = ForeignKey('self', null=True, blank=True, on_delete=CASCADE)
    path = PathField(order_by=['century', 'last_name', 'first_name'])

    class Meta:
        ordering = ['path']
        indexes = [
            *PathField.get_indexes('person', 'path'),
        ]
